from __future__ import annotations

import importlib
import uuid
from typing import TYPE_CHECKING

import numpy as np

import gamspy as gp
import gamspy._algebra.expression as expression
import gamspy.formulations.utils as utils
from gamspy.exceptions import ValidationError
from gamspy.formulations.ml.decision_tree_struct import DecisionTreeStruct

if TYPE_CHECKING:
    from sklearn.tree import DecisionTreeRegressor


class RegressionTree:
    """
    Formulation generator for Regression Trees in GAMSPy.

    Parameters
    ----------
    container : Container
        Container that will contain the new variable and equations.
    regressor: DecisionTreeRegressor | DecisionTreeStruct
        - A fitted `sklearn.tree.DecisionTreeRegressor` instance.
        - If `sklearn.tree.DecisionTreeRegressor` is not utilized, the fitted tree information
          can be supplied via the `DecisionTreeStruct` dataclass, which represents the same components as those in `sklearn.tree`.
          See :meth:`DecisionTreeStruct <gamspy.formulations.DecisionTreeStruct>` for details on required attributes.

    name_prefix : str | None
        Prefix for generated GAMSPy symbols, by default None which means
        random prefix. Using the same name_prefix in different formulations causes name
        conflicts. Do not use the same name_prefix again.

    Examples
    --------
    >>> import gamspy as gp
    >>> import numpy as np
    >>> from gamspy.math import dim
    >>> np.random.seed(42)
    >>> m = gp.Container()
    >>> in_data = np.random.randint(0, 10, size=(5, 2))
    >>> out_data = np.random.randint(1, 3, size=(5, 1))
    >>> tree_attribute = {
    ...    "children_left": np.array([1, 2, -1, -1, -1]),
    ...    "children_right": np.array([4, 3, -1, -1, -1]),
    ...    "feature": np.array([0, 1, -2, -2, -2]),
    ...    "threshold": np.array([5.5, 4.5, -2.0, -2.0, -2.0]),
    ...    "value": np.array([[15.6], [11.25], [10.0], [15.0], [33.0]]),
    ...    "capacity": 5,
    ...    "n_features": 2,
    ... }
    >>> tree = gp.formulations.DecisionTreeStruct(**tree_attribute)
    >>> dt_model = gp.formulations.RegressionTree(m, tree)
    >>> x = gp.Variable(m, "x", domain=dim((5, 2)), type="positive")
    >>> x.up[:, :] = 10
    >>> y, eqns = dt_model(x)
    >>> set_of_samples = y.domain[0]
    >>> set_of_samples.name
    'DenseDim5_1'

    """

    def __init__(
        self,
        container: gp.Container,
        regressor: DecisionTreeRegressor | DecisionTreeStruct,
        name_prefix: str | None = None,
    ):
        if not isinstance(container, gp.Container):
            raise ValidationError(f"{container} is not a gp.Container.")

        if isinstance(regressor, DecisionTreeStruct):
            self._initialize_from_decision_tree_struct(regressor=regressor)
        else:
            try:
                sklearn_tree = importlib.import_module("sklearn.tree")
                if isinstance(regressor, sklearn_tree.DecisionTreeRegressor):
                    self._initialize_from_sklearn(regressor=regressor)
                else:
                    raise ValidationError(
                        f"{regressor} must be an instance of either >sklearn.tree.DecisionTreeRegressor< or >DecisionTreeStruct<"
                    )
            except ModuleNotFoundError:
                raise ValidationError(">sklearn.tree< module not found.") from None

        self._check_tree()

        self.container = container

        if name_prefix is None:
            name_prefix = str(uuid.uuid4()).split("-")[0]

        self._name_prefix = name_prefix

    def _initialize_from_sklearn(self, regressor: DecisionTreeRegressor):
        """
        Initializes the tree attributes from a fitted scikit-learn DecisionTreeRegressor.
        """
        _trained = hasattr(regressor, "tree_") and regressor.tree_ is not None

        if not _trained:
            raise ValidationError(
                f"{regressor} is not fitted or tree_ attribute is missing."
            )

        self.children_left = regressor.tree_.children_left
        self.children_right = regressor.tree_.children_right
        self.feature = regressor.tree_.feature
        self.threshold = regressor.tree_.threshold
        self.value = regressor.tree_.value[:, :, 0]
        self.capacity = regressor.tree_.capacity
        self.n_features = regressor.tree_.n_features

    def _initialize_from_decision_tree_struct(self, regressor: DecisionTreeStruct):
        """
        Initializes the tree attributes from DecisionTreeStruct.
        """
        self.children_left = regressor.children_left
        self.children_right = regressor.children_right
        self.feature = regressor.feature
        self.threshold = regressor.threshold
        self.value = regressor.value
        self.capacity = regressor.capacity
        self.n_features = regressor.n_features

    def _check_tree(self):
        """
        Validates that the core tree attributes are populated and have valid basic properties.

        Raises:
            ValidationError: If any essential attribute is empty or invalid.
        """
        if self.children_left.size == 0:
            raise ValidationError("Attribute 'children_left' is empty.")
        if self.children_right.size == 0:
            raise ValidationError("Attribute 'children_right' is empty.")
        if self.feature.size == 0:
            raise ValidationError("Attribute 'feature' is empty.")
        if self.threshold.size == 0:
            raise ValidationError("Attribute 'threshold' is empty.")
        if self.value.size == 0:
            raise ValidationError("Attribute 'value' is empty.")
        if self.n_features <= 0:
            raise ValidationError("'n_features' must be a positive integer.")
        if self.capacity <= 0:
            raise ValidationError("'capacity' must be a positive integer.")

    def _node_bounds(self):
        """Traverse the tree using DFS and extract bound information for each node."""
        node_lb = np.empty((self.n_features, self.capacity))
        node_lb.fill(-np.inf)
        node_ub = np.empty((self.n_features, self.capacity))
        node_ub.fill(np.inf)

        stack = [0]

        while stack:
            node = stack.pop()
            left = self.children_left[node]
            if left < 0:
                continue
            right = self.children_right[node]

            node_lb[:, left] = node_lb[:, node]
            node_ub[:, left] = node_ub[:, node]
            node_lb[:, right] = node_lb[:, node]
            node_ub[:, right] = node_ub[:, node]

            node_ub[self.feature[node], left] = self.threshold[node]
            node_lb[self.feature[node], right] = self.threshold[node]

            stack.extend((right, left))

        return node_lb, node_ub

    def _yield_call(
        self,
        input: gp.Parameter | gp.Variable,
        M: float | None = None,
    ):
        leafs = self.children_left < 0
        leafs = leafs.nonzero()[0]
        nleafs = len(leafs)
        output_dim = self.value.shape[-1]
        node_lb, node_ub = self._node_bounds()

        if not isinstance(input, (gp.Variable, gp.Parameter)):
            raise ValidationError(
                "Input must be of either type gp.Parameter | gp.Variable"
            )

        if len(input.domain) != 2:
            raise ValidationError(
                f"input expected to be in shape (n_samples, {self.n_features})"
            )

        if len(input.domain[-1]) != self.n_features:
            raise ValidationError("number of features do not match")

        if M is not None and not isinstance(M, (float, int)):
            raise ValidationError("M can either be of type float or int")

        set_of_samples: gp.Set = input.domain[0]
        set_of_features: gp.Set = input.domain[-1]

        [
            set_of_output_dim,
            set_of_lb_ub,
            set_of_leafs,
        ] = gp.math._generate_dims(self.container, dims=[output_dim, 2, nleafs])

        [
            set_of_samples,
            set_of_features,
            set_of_output_dim,
            set_of_lb_ub,
            set_of_leafs,
        ] = utils._next_domains(
            [
                set_of_samples,
                set_of_features,
                set_of_output_dim,
                set_of_lb_ub,
                set_of_leafs,
            ],
            [],
        )

        lb = ge = "0"
        ub = le = "1"
        cons_type = set_of_lb_ub

        feat_par = gp.Parameter._constructor_bypass(
            self.container,
            name=utils._generate_name("p", self._name_prefix, "feat_par"),
            domain=[set_of_features, set_of_leafs, set_of_lb_ub],
        )

        out = gp.Variable._constructor_bypass(
            self.container,
            name=utils._generate_name("v", self._name_prefix, "output"),
            domain=[set_of_samples, set_of_output_dim],
        )

        yield out

        feat_vars = gp.Variable._constructor_bypass(
            self.container,
            name=utils._generate_name("v", self._name_prefix, "feature"),
            domain=[set_of_samples, set_of_features],
        )

        indicator_vars = gp.Variable._constructor_bypass(
            self.container,
            name=utils._generate_name("v", self._name_prefix, "indicator"),
            type="binary",
            domain=[set_of_samples, set_of_leafs],
            description="indicator variable for each leaf for each sample",
        )

        assign_one_output = gp.Equation._constructor_bypass(
            self.container,
            name=utils._generate_name("e", self._name_prefix, "assign_one_output"),
            domain=[set_of_samples],
            description="Activate only one leaf per sample",
        )

        out_link = gp.Parameter._constructor_bypass(
            self.container,
            name=utils._generate_name("p", self._name_prefix, "predicted_value"),
            domain=[set_of_leafs, set_of_output_dim],
        )

        link_indicator_output = gp.Equation._constructor_bypass(
            self.container,
            name=utils._generate_name("e", self._name_prefix, "link_indicator_output"),
            domain=[set_of_samples, set_of_output_dim],
            description="Link the indicator variable to the predicted value of the decision tree",
        )

        max_out = gp.Parameter._constructor_bypass(
            self.container,
            name=utils._generate_name("p", self._name_prefix, "max_out"),
            domain=[set_of_output_dim],
        )
        min_out = gp.Parameter._constructor_bypass(
            self.container,
            name=utils._generate_name("p", self._name_prefix, "min_out"),
            domain=[set_of_output_dim],
        )

        ub_output = gp.Equation._constructor_bypass(
            self.container,
            name=utils._generate_name("e", self._name_prefix, "ub_output"),
            domain=[set_of_samples, set_of_output_dim],
            description="Output cannot be more than the maximum of predicted value",
        )

        lb_output = gp.Equation._constructor_bypass(
            self.container,
            name=utils._generate_name("e", self._name_prefix, "lb_output"),
            domain=[set_of_samples, set_of_output_dim],
            description="Output cannot be less than the minimum of predicted value",
        )

        uni_domain = [set_of_samples, set_of_features, set_of_leafs]

        s = gp.Set._constructor_bypass(
            self.container,
            name=utils._generate_name("s", self._name_prefix, "subset_of_paths"),
            description="Dynamic subset of possible paths",
            domain=uni_domain,
        )

        feat_thresh = gp.Parameter._constructor_bypass(
            self.container,
            name=utils._generate_name("p", self._name_prefix, "feature_threshold"),
            description="feature splitting value",
            domain=uni_domain + [cons_type],
        )

        _bound_big_m = gp.Parameter._constructor_bypass(
            self.container,
            name=utils._generate_name("p", self._name_prefix, "bound_big_m"),
            description="bound value for big-M",
            domain=uni_domain + [cons_type],
        )

        ge_cons = gp.Equation._constructor_bypass(
            self.container,
            name=utils._generate_name(
                "e", self._name_prefix, "link_indicator_feature_ge"
            ),
            domain=uni_domain,
            description="Link the indicator variable with the feature which is Lower bounded using a big-M constraint",
        )

        le_cons = gp.Equation._constructor_bypass(
            self.container,
            name=utils._generate_name(
                "e", self._name_prefix, "link_indicator_feature_le"
            ),
            domain=uni_domain,
            description="Link the indicator variable with the feature which is Upper bounded using a big-M constraint",
        )

        if isinstance(input, gp.Variable):
            feat_vars = input
        else:
            self.container._add_statement(
                expression.Expression(feat_vars.fx[...], "=", input[...])
            )

        definition = expression.Expression(
            assign_one_output[set_of_samples],
            "..",
            (gp.Sum(set_of_leafs, indicator_vars[set_of_samples, set_of_leafs]) == 1),
        )
        self.container._add_statement(definition)
        assign_one_output._definition = definition

        idx, jdx = np.indices((nleafs, output_dim), dtype=int)
        mapped_values = self.value[leafs[:, None], jdx]

        definition = expression.Expression(
            link_indicator_output[set_of_samples, set_of_output_dim],
            "..",
            (
                gp.Sum(
                    set_of_leafs,
                    out_link[set_of_leafs, set_of_output_dim]
                    * indicator_vars[set_of_samples, set_of_leafs],
                )
                == out
            ),
        )
        self.container._add_statement(definition)
        link_indicator_output._definition = definition

        definition = expression.Expression(ub_output[...], "..", out <= max_out)
        self.container._add_statement(definition)
        ub_output._definition = definition

        definition = expression.Expression(lb_output[...], "..", out >= min_out)
        lb_output._definition = definition
        self.container._add_statement(definition)

        set_records_dict = {
            out_link: [
                (int(i), int(j), v)
                for i, j, v in np.stack((idx, jdx, mapped_values), axis=-1).reshape(
                    -1, 3
                )
            ],
            feat_par: np.stack([node_lb, node_ub], axis=-1)[:, leafs, :],
            max_out: np.max(self.value[leafs, :], axis=0),
            min_out: np.min(self.value[leafs, :], axis=0),
        }

        yield [set_of_output_dim, set_records_dict]

        ### This generates the set of possible paths given the input data
        mask = (feat_vars.up[...] >= feat_par[..., lb]) & (
            feat_vars.lo[...] <= feat_par[..., ub]
        )
        mask_lo = (
            (feat_vars.lo[...] < feat_par[..., lb])
            & mask
            & (feat_par[..., lb] > -np.inf)
        )
        mask_up = (
            (feat_vars.up[...] > feat_par[..., ub])
            & mask
            & (feat_par[..., ub] < np.inf)
        )
        s[...].where[mask_lo | mask_up] = True

        self.container._add_statement(
            expression.Expression(
                feat_thresh[..., ge].where[mask_lo],
                "=",
                feat_par[..., lb],
            )
        )
        self.container._add_statement(
            expression.Expression(
                feat_thresh[..., le].where[mask_up],
                "=",
                feat_par[..., ub],
            )
        )
        self.container._add_statement(
            expression.Expression(
                indicator_vars.fx[set_of_samples, set_of_leafs].where[
                    gp.Sum(set_of_features, ~mask)
                ],
                "=",
                0,
            )
        )
        self.container._synch_with_gams()

        self.container._add_statement(
            expression.Expression(
                _bound_big_m[s[uni_domain], ge].where[mask_lo],
                "=",
                (
                    M
                    if M
                    else (
                        feat_par[set_of_features, set_of_leafs, lb]
                        - feat_vars.lo[set_of_samples, set_of_features]
                    )
                ),
            )
        )

        self.container._add_statement(
            expression.Expression(
                _bound_big_m[s[uni_domain], le].where[mask_up],
                "=",
                (
                    M
                    if M
                    else (
                        feat_vars.up[set_of_samples, set_of_features]
                        - feat_par[set_of_features, set_of_leafs, ub]
                    )
                ),
            )
        )

        self.container._add_statement(
            expression.Expression(
                _bound_big_m[...].where[gp.math.abs(_bound_big_m) == np.inf],
                "=",
                1e10,
            )
        )

        definition = expression.Expression(
            ge_cons[uni_domain].where[(feat_thresh[..., ge] != 0) & s[uni_domain]],
            "..",
            feat_vars
            >= feat_thresh[..., ge] - _bound_big_m[..., ge] * (1 - indicator_vars),
        )
        self.container._add_statement(definition)
        ge_cons._definition = definition

        le_cons[uni_domain].where[(feat_thresh[..., le] != 0) & s[uni_domain]] = (
            feat_vars
            <= feat_thresh[..., le] + _bound_big_m[..., le] * (1 - indicator_vars)
        )

        eqns = [
            assign_one_output,
            link_indicator_output,
            ub_output,
            lb_output,
            ge_cons,
            le_cons,
        ]
        yield eqns

    def __call__(
        self,
        input: gp.Parameter | gp.Variable,
        M: float | None = None,
    ) -> tuple[gp.Variable, list[gp.Equation]]:
        """
        Generate output variable and equations required for embedding the regression tree.

        Parameters
        ----------
        input : gp.Parameter | gp.Variable
                input for the regression tree, must be in shape (sample_size, number_of_features)
        M : float
            value for the big_M. By default, infer the value using the available bounds for variables.
            If the variable is unbounded, then default to 1e10.
        """

        generator = self._yield_call(input, M)

        output = next(generator)
        _, set_values = next(generator)
        input.container.setRecords(set_values)
        eqs = next(generator)

        return output, eqs



from __future__ import annotations

import typer
from rich import print
from rich.console import Console
from rich.table import Table

import gamspy.utils as utils
from gamspy.exceptions import ValidationError

console = Console()
app = typer.Typer(
    rich_markup_mode="rich",
    short_help="To list solvers.",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy list solvers --all | gamspy list solvers --defaults",
    context_settings={"help_option_names": ["-h", "--help"]},
)


@app.command()
def solvers(
    all: bool = typer.Option(False, "--all", "-a", help="Shows all available solvers."),
    installables: bool = typer.Option(
        False,
        "--installables",
        "-i",
        help="Shows solvers that can be installed.",
    ),
    defaults: bool = typer.Option(
        False, "--defaults", "-d", help="Shows default solvers."
    ),
) -> None:
    try:
        import gamspy_base
    except ModuleNotFoundError as e:
        raise ValidationError(
            "You must install gamspy_base to use this command!"
        ) from e

    capabilities = utils.getSolverCapabilities(gamspy_base.directory)
    if all:
        solvers = utils.getAvailableSolvers()
        print("[bold]Available Solvers[/bold]")
        print("=" * 17)
        print(", ".join(solvers))
        print(
            "\n[bold]Model types that can be solved with the installed solvers[/bold]\n"
        )
        table = Table("Solver", "Problem Types")
        for solver in solvers:
            try:
                table.add_row(solver, ", ".join(capabilities[solver]))
            except KeyError:
                ...
        console.print(table)

        print(
            "[bold]Full list can be found here[/bold]: https://www.gams.com/latest/docs/S_MAIN.html#SOLVERS_MODEL_TYPES"
        )
    elif installables:
        installable_solvers = utils.getInstallableSolvers(gamspy_base.directory)
        console.print(", ".join(installable_solvers))
    elif defaults:
        default_solvers = utils.getDefaultSolvers(gamspy_base.directory)
        table = Table("Problem", "Solver")
        for problem in default_solvers:
            try:
                table.add_row(problem, default_solvers[problem])
            except KeyError:
                ...

        console.print(table)
    else:
        solvers = utils.getInstalledSolvers(gamspy_base.directory)
        print("[bold]Installed Solvers[/bold]")
        print("=" * 17)
        print(", ".join(solvers))

        print(
            "\n[bold]Model types that can be solved with the installed solvers[/bold]"
        )
        print("=" * 57)
        table = Table("Solver", "Problem Types")
        for solver in solvers:
            try:
                table.add_row(solver, ", ".join(capabilities[solver]))
            except KeyError:
                ...

        console.print(table)


if __name__ == "__main__":
    app()



from __future__ import annotations

import inspect
import logging
import os
import threading
from collections.abc import Sequence
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar

from gams.core.gdx import GMS_UEL_IDENT_SIZE

import gamspy as gp
import gamspy._algebra.expression as expression
import gamspy._algebra.operation as operation
import gamspy._miro as miro
import gamspy._symbols.implicits as implicits
import gamspy._validation as validation
import gamspy.utils as utils
from gamspy._backend.backend import backend_factory
from gamspy._convert import (
    GamsConverter,
    LatexConverter,
    get_convert_solver_options,
)
from gamspy._model_instance import ModelInstance
from gamspy._options import (
    EXECUTION_OPTIONS,
    MODEL_ATTR_OPTION_MAP,
    ConvertOptions,
    FreezeOptions,
    Options,
)
from gamspy.exceptions import GamspyException, ValidationError

if TYPE_CHECKING:
    from typing import Literal, TextIO

    import pandas as pd

    from gamspy import Container, Equation, Parameter, Variable
    from gamspy._algebra.expression import Expression
    from gamspy._algebra.operation import Operation
    from gamspy._backend.engine import EngineClient
    from gamspy._backend.neos import NeosClient
    from gamspy._symbols.implicits import ImplicitParameter, ImplicitVariable
    from gamspy._symbols.symbol import Symbol
    from gamspy.math import MathOp

GMS_MAX_LINE_LENGTH = 80000
MAX_MODEL_DECLARATION_LENGTH = 75
MAX_SYMBOL_NAME_LENGTH = GMS_UEL_IDENT_SIZE - 1
MAX_MATCHING_LENGTH = (
    MAX_SYMBOL_NAME_LENGTH + MAX_SYMBOL_NAME_LENGTH + MAX_MODEL_DECLARATION_LENGTH + 1
)
MAX_NUM_MODEL_ELEMS = int(GMS_MAX_LINE_LENGTH / MAX_MATCHING_LENGTH)
IS_MIRO_INIT = os.getenv("MIRO", False)

logger = logging.getLogger("MODEL")
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)
formatter = logging.Formatter("[%(name)s - %(levelname)s] %(message)s")
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)


class Problem(Enum):
    """An enumeration for problem all problem types"""

    LP = "LP"
    "Linear Programming"

    NLP = "NLP"
    """Non-Linear Programming"""

    QCP = "QCP"
    """Quadratically Constrained Programs"""

    DNLP = "DNLP"
    """Nonlinear Programming with Discontinuous Derivatives"""

    MIP = "MIP"
    """Mixed Integer Programming"""

    RMIP = "RMIP"
    """Relaxed Mixed Integer Program"""

    MINLP = "MINLP"
    """Mixed Integer Nonlinear Program"""

    RMINLP = "RMINLP"
    """Relaxed Mixed Integer Nonlinear Program"""

    MIQCP = "MIQCP"
    """Mixed Integer Quadratically Constrained Program"""

    RMIQCP = "RMIQCP"
    """Relaxed Mixed Integer Quadratically Constrained Program"""

    MCP = "MCP"
    """Mixed Complementarity Problem"""

    CNS = "CNS"
    """Constrained Nonlinear System"""

    MPEC = "MPEC"
    """Mathematical Programs with Equilibrium Constraints"""

    RMPEC = "RMPEC"
    """Relaxed Mathematical Program with Equilibrium Constraints"""

    EMP = "EMP"
    """Extended Mathematical Program"""

    MPSGE = "MPSGE"
    """General Equilibrium"""

    @classmethod
    def values(cls):
        """Convenience function to return all values of enum"""
        return list(cls._value2member_map_.keys())

    def __str__(self) -> str:
        return self.value


class Sense(Enum):
    """An enumeration for sense types"""

    MIN = "MIN"
    """Minimize the objective."""

    MAX = "MAX"
    """Maximize the objective."""

    FEASIBILITY = "FEASIBILITY"
    """Assess feasibility."""

    @classmethod
    def values(cls):
        """Convenience function to return all values of enum"""
        return list(cls._value2member_map_.keys())

    def __str__(self) -> str:
        return self.value


class ModelStatus(Enum):
    """An enumeration for model status types"""

    OptimalGlobal = 1
    """The solution is optimal, that is, it is feasible (within tolerances) and it has been proven that no other feasible solution with better objective value exists."""

    OptimalLocal = 2
    """A local optimum for an NLP has been found. That is, a solution that is feasible (within tolerances) and it has been proven that there exists a neighborhood of this solution in which no other feasible solution with better objective value exists."""

    Unbounded = 3
    """The solution is unbounded. This message is reliable if the problem is linear, but occasionally it appears for difficult nonlinear problems that are not truly unbounded, but that lack some strategically placed bounds to limit the variables to sensible values."""

    InfeasibleGlobal = 4
    """The problem has been proven to be infeasible. If this was not intended, something is probably misspecified in the logic or the data."""

    InfeasibleLocal = 5
    """No feasible point could be found for the NLP problem from the given starting point. It does not necessarily mean that no feasible point exists."""

    InfeasibleIntermed = 6
    """The current solution is not feasible, but the solver stopped, either because of a limit (for example, iteration or resource) or because of some sort of difficulty. The solver status will give more information."""

    Feasible = 7
    """A feasible solution to a problem without discrete variables has been found."""

    Integer = 8
    """A feasible solution to a problem with discrete variables has been found."""

    NonIntegerIntermed = 9
    """An incomplete solution to a problem with discrete variables. A feasible solution has not yet been found."""

    IntegerInfeasible = 10
    """It has been proven that there is no feasible solution to a problem with discrete variables."""

    LicenseError = 11
    """The solver cannot find the appropriate license key needed to use a specific subsolver."""

    ErrorUnknown = 12
    """After a solver error the model status is unknown."""

    ErrorNoSolution = 13
    """An error occurred and no solution has been returned. No solution will be returned to GAMS because of errors in the solution process."""

    NoSolutionReturned = 14
    """A solution is not expected for this solve. For example, the CONVERT solver only reformats the model but does not give a solution."""

    SolvedUnique = 15
    """Indicates the solution returned is unique, i.e. no other solution exists. Used for CNS models. Examples where this status could be returned include non-singular linear models, triangular models with constant non-zero elements on the diagonal, and triangular models where the functions are monotone in the variable on the diagonal."""

    Solved = 16
    """Indicates the model has been solved: used for CNS models. The solution might or might not be unique. If the solver uses status SOLVED SINGULAR wherever possible then this status implies that the Jacobian is non-singular, i.e. that the solution is at least locally unique."""

    SolvedSingular = 17
    """Indicates the CNS model has been solved, but the Jacobian is singular at the solution. This can indicate that other solutions exist, either along a line (for linear models) or a curve (for nonlinear models) including the solution returned."""

    UnboundedNoSolution = 18
    """The model is unbounded and no solution can be provided."""

    InfeasibleNoSolution = 19
    """The model is infeasible and no solution can be provided."""


class SolveStatus(Enum):
    """An enumeration for solve status types"""

    NormalCompletion = 1
    """The solver terminated in a normal way."""

    IterationInterrupt = 2
    """The solver was interrupted because it used too many iterations. The option `iteration_limit` may be used to increase the iteration limit if everything seems normal."""

    ResourceInterrupt = 3
    """The solver was interrupted because it used too much time. The option `time_limit` may be used to increase the time limit if everything seems normal."""

    TerminatedBySolver = 4
    """The solver encountered some difficulty and was unable to continue."""

    EvaluationInterrupt = 5
    """Too many evaluations of nonlinear terms at undefined values. We recommend to use variable bounds to prevent forbidden operations, such as division by zero. The rows in which the errors occur are listed just before the solution."""

    CapabilityError = 6
    """The solver does not have the capability required by the model. For example, some solvers do not support certain types of discrete variables or support a more limited set of functions than other solvers."""

    LicenseError = 7
    """The solver cannot find the appropriate license key needed to use a specific subsolver."""

    UserInterrupt = 8
    """The user has sent a signal to interrupt the solver."""

    SetupError = 9
    """The solver encountered a fatal failure during problem set-up time."""

    SolverError = 10
    """The solver encountered a fatal error."""

    InternalError = 11
    """The solver encountered an internal fatal error."""

    Skipped = 12
    """The entire solve step has been skipped."""

    SystemError = 13
    """This indicates a completely unknown or unexpected error condition."""


class FileFormat(Enum):
    """An enumeration for file format types"""

    AMPL = "ampl.mod"
    """AMPL input format."""

    AMPLNL = "ampl.nl"
    """AMPL nl format."""

    CPLEXLP = "cplex.lp"
    """CPLEX LP format."""

    CPLEXMPS = "cplex.mps"
    """CPLEX MPS format."""

    GAMSDict = "dict.txt"
    """GAMS dictionary format."""

    GAMSDictMap = "dictmap.gdx"
    """GAMS dictionary map format."""

    GAMSJacobian = "jacobian.gms"
    """Jacobian in GAMS."""

    GAMSPyJacobian = "jacobian.py"
    """Jacobian in GAMSPy."""

    GDXJacobian = "jacobian.gdx"
    """GDX file with model data incl. Jacobian and Hessian evaluated at current point."""

    FileList = "files.txt"
    """List of file formats generated."""

    FixedMPS = "fixed.mps"
    """Fixed format mps file."""

    GAMS = "gams.gms"
    """GAMS scalar model."""

    JuMP = "jump.jl"
    """JuMP scalar model."""

    LINGO = "lingo.lng"
    """Lingo format."""

    OSiL = "osil.xml"
    """Optimization Services instance Language (OSiL) format."""

    Pyomo = "pyomo.py"
    """Pyomo concrete scalar model."""


INTERRUPT_STATUS = [
    SolveStatus.IterationInterrupt,
    SolveStatus.ResourceInterrupt,
    SolveStatus.EvaluationInterrupt,
    SolveStatus.UserInterrupt,
    SolveStatus.TerminatedBySolver,
]

ERROR_STATUS = {
    SolveStatus.CapabilityError: "The solver does not have the capability required by the model.",
    SolveStatus.LicenseError: "The solver cannot find the appropriate license key needed to use a specific subsolver.",
    SolveStatus.SetupError: "The solver encountered a fatal failure during problem set-up time.",
    SolveStatus.SolverError: "The solver encountered a fatal error.",
    SolveStatus.InternalError: "The solver encountered an internal fatal error.",
    SolveStatus.SystemError: "This indicates a completely unknown or unexpected error condition.",
}


# GAMS name -> GAMSPy name
ATTRIBUTE_MAP = {
    "domUsd": "_num_domain_violations",
    "etAlg": "_algorithm_time",
    "etSolve": "_total_solve_time",
    "etSolver": "_total_solver_time",
    "iterUsd": "_num_iterations",
    "marginals": "_marginals",
    "maxInfes": "_max_infeasibility",
    "meanInfes": "_mean_infeasibility",
    "modelStat": "_status",
    "nodUsd": "_num_nodes_used",
    "number": "_solve_number",
    "numDepnd": "_num_dependencies",
    "numDVar": "_num_discrete_variables",
    "numEqu": "_num_equations",
    "numInfes": "_num_infeasibilities",
    "numNLIns": "_num_nonlinear_insts",
    "numNLNZ": "_num_nonlinear_zeros",
    "numNOpt": "_num_nonoptimalities",
    "numNZ": "_num_nonzeros",
    "numRedef": "_num_mcp_redefinitions",
    "numVar": "_num_variables",
    "numVarProj": "_num_bound_projections",
    "objEst": "_objective_estimation",
    "objVal": "_objective_value",
    "procUsed": "_used_model_type",
    "resGen": "_model_generation_time",
    "resUsd": "_solve_model_time",
    "solveStat": "_solve_status",
    "sumInfes": "_sum_infeasibilities",
    "sysVer": "_solver_version",
}


class Model:
    """
    Represents a list of equations to be solved.

    Parameters
    ----------
    container : Container
        Container of the model.
    name : str, optional
        Name of the model. Name is autogenerated by default.
    description : str, optional
        Description of the model.
    equations : Sequence[Equation]
        Sequence of Equation objects.
    problem : Problem | str, optional
        'LP', 'NLP', 'QCP', 'DNLP', 'MIP', 'RMIP', 'MINLP', 'RMINLP', 'MIQCP', 'RMIQCP', 'MCP', 'CNS', 'MPEC', 'RMPEC', 'EMP', or 'MPSGE',
        by default Problem.LP.
    sense : Sense | str, optional
        "MIN", "MAX", or "FEASIBILITY". By default, Sense.FEASIBILITY
    objective : Variable | Expression, optional
        Objective variable to minimize or maximize or objective itself.
    matches : dict[Equation | Sequence[Equation], Variable | Sequence[Variable]], optional
        Equation - Variable matches for MCP models.
    limited_variables : Sequence[ImplicitVariable], optional
        Allows limiting the domain of variables used in a model.
    external_module: str, optional
        The name of the external module in which the external equations are implemented

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> v = gp.Variable(m, "v")
    >>> e = gp.Equation(m, "e", definition= v == 5)
    >>> my_model = gp.Model(m, "my_model", problem="LP", equations=[e])

    """

    # Prefix for auto-generated symbols
    _generate_prefix = "autogenerated_"
    _autogen_symbols: ClassVar[set[str]] = set()

    def __init__(
        self,
        container: Container | None = None,
        name: str | None = None,
        description: str = "",
        problem: Problem | str = Problem.MIP,
        equations: Sequence[Equation] | None = None,
        sense: Sense | str = Sense.FEASIBILITY,
        objective: Variable | Expression | Operation | None = None,
        matches: dict[
            Equation | Sequence[Equation],
            Variable | Sequence[Variable],
        ]
        | None = None,
        limited_variables: Sequence[ImplicitVariable] | None = None,
        external_module: str | None = None,
    ):
        self._auto_id = "m" + utils._get_unique_name()
        if equations is None:
            equations = []

        self.description = description

        if container is not None:
            self.container = container
        else:
            self.container = gp._ctx_managers[(os.getpid(), threading.get_native_id())]

        assert self.container is not None

        if name is not None:
            name = validation.validate_name(name)
            self.name = validation.validate_model_name(name)
        else:
            self.name = self.container._get_symbol_name(prefix="m")

        self._matches = matches
        self.problem, self.sense = validation.validate_model(
            equations, matches, problem, sense
        )
        self.equations = list(equations)
        self._objective = objective
        self._objective_variable = self._set_objective_variable(objective)
        if (
            self._objective_variable is not None
            and self._objective_variable.type.lower() != "free"
        ):
            raise ValidationError(
                f"Objective variable `{self._objective_variable}` must be a free variable"
            )

        if limited_variables is not None and not isinstance(
            limited_variables, Sequence
        ):
            raise ValidationError(
                f"`limited_variables must be an Iterable of ImplicitVariable objects but found {limited_variables}`"
            )

        self._limited_variables: Sequence[ImplicitVariable] | None = limited_variables

        if not self.equations and not self._matches:
            raise ValidationError("Model requires at least one equation.")

        self._external_module_file: str | None = None
        self._external_module: str | None = None

        if external_module is not None:
            self.external_module = external_module
        else:
            # To avoid adding it twice
            self.container._add_statement(self)

        # allow freezing
        self._is_frozen: bool = False

        # Attributes
        self._num_domain_violations: float | None = None
        self._algorithm_time: float | None = None
        self._total_solve_time: float | None = None
        self._total_solver_time: float | None = None
        self._num_iterations: float | None = None
        self._marginals: float | None = None
        self._max_infeasibility: float | None = None
        self._mean_infeasibility: float | None = None
        self._status: ModelStatus | None = None
        self._num_nodes_used: float | None = None
        self._solve_number: float | None = None
        self._num_dependencies: float | None = None
        self._num_discrete_variables: float | None = None
        self._num_equations: float | None = None
        self._num_infeasibilities: float | None = None
        self._num_nonlinear_insts: float | None = None
        self._num_nonlinear_zeros: float | None = None
        self._num_nonoptimalities: float | None = None
        self._num_nonzeros: float | None = None
        self._num_mcp_redefinitions: float | None = None
        self._num_variables: float | None = None
        self._num_bound_projections: float | None = None
        self._objective_estimation: float | None = None
        self._objective_value: float | None = None
        self._used_model_type: float | None = None
        self._model_generation_time: float | None = None
        self._solve_model_time: float | None = None
        self._sum_infeasibilities: float | None = None
        self._solve_status: SolveStatus | None = None
        self._solver_version: float | None = None

        self.container.models.update({self.name: self})
        self.container._synch_with_gams()

    def _serialize(self) -> dict:
        info: dict[str, Any] = {
            "name": self.name,
            "problem": str(self.problem),
            "sense": str(self.sense),
        }

        # equations
        equations = [equation.name for equation in self.equations]
        info["equations"] = equations

        # matches
        if self._matches is not None:
            matches: dict = {}
            for key, value in self._matches.items():
                if isinstance(key, gp.Equation):
                    if isinstance(value, gp.Variable):
                        matches[key.name] = value.name
                    else:
                        matches[key.name] = (variable.name for variable in value)
                else:
                    assert isinstance(value, gp.Variable)
                    for equation in key:
                        matches[equation.name] = value.name

            info["_matches"] = matches

        # objective variable
        if self._objective_variable is not None:
            info["_objective_variable"] = self._objective_variable.name

        # attributes
        for attr_name in ATTRIBUTE_MAP.values():
            if attr_name in ("_status", "_solve_status"):
                attr_value = getattr(self, attr_name, None)
                if attr_value is not None:
                    attr_value = attr_value.value

                info[attr_name] = attr_value
            else:
                info[attr_name] = getattr(self, attr_name, None)

        return info

    def __repr__(self) -> str:
        return f"Model(name='{self.name}', problem='{self.problem!s}', equations={self.equations}, sense='{self.sense!s}', objective={self._objective_variable}, matches={self._matches}, limited_variables={self._limited_variables}"

    def __str__(self) -> str:
        return (
            f"Model {self.name}:\n  Problem Type: {self.problem}\n  Sense:"
            f" {self.sense}\n  Equations: {self.equations}"
        )

    @property
    def num_domain_violations(self) -> float | None:
        """
        Number of domain violations.

        Returns
        -------
        float | None
        """
        return self._num_domain_violations

    @property
    def algorithm_time(self) -> float | None:
        """
        Solver dependent timing information. This attribute was intended to allow
        solvers to return the elapsed time used by the solve algorithm without
        including any model generation, communication, or setup time. However,
        solvers are free to adapt this convention and return time-related information
        (but not necessarily elapsed time) for executing the solve algorithm. Please
        inspect your solver manual for the actual meaning of the value returned in
        this attribute.

        Returns
        -------
        float | None
        """
        return self._algorithm_time

    @property
    def total_solve_time(self) -> float | None:
        """
        Elapsed time it took to execute a solve statement in total. This model
        attribute returns the elapsed time it took to execute a solve statement in total.
        This time includes the model generation time, the time to read and write files,
        the time to create the solution report and the time taken by the actual solve.
        The time is expressed in seconds of wall-clock time.

        Returns
        -------
        float | None
        """
        return self._total_solve_time

    @property
    def total_solver_time(self) -> float | None:
        """
        Elapsed time taken by the solver only. This model attribute returns the elapsed
        time taken by the solver only. This does not include the GAMS model generation time
        and the time taken to report and load the solution back into the GAMS database.
        The time is expressed in seconds of wall-clock time.

        Returns
        -------
        float | None
        """
        return self._total_solver_time

    @property
    def num_iterations(self) -> float | None:
        """
        Number of iterations used.

        Returns
        -------
        float | None
        """
        return self._num_iterations

    @property
    def marginals(self) -> float | None:
        """
        Indicates whether there are marginals.

        Returns
        -------
        float | None
        """
        return self._marginals

    @property
    def max_infeasibility(self) -> float | None:
        """
        Maximum of infeasibilities

        Returns
        -------
        float | None
        """
        return self._max_infeasibility

    @property
    def mean_infeasibility(self) -> float | None:
        """
        Mean of infeasibilities

        Returns
        -------
        float | None
        """
        return self._mean_infeasibility

    @property
    def status(self) -> ModelStatus | None:
        """
        Model status after solve.

        Returns
        -------
        ModelStatus | None
        """
        return self._status

    @property
    def num_nodes_used(self) -> float | None:
        """
        Number of nodes used by the MIP solver.

        Returns
        -------
        float | None
        """
        return self._num_nodes_used

    @property
    def solve_number(self) -> float | None:
        """
        Number of the last solve.

        Returns
        -------
        float | None
        """
        return self._solve_number

    @property
    def num_dependencies(self) -> float | None:
        """
        Number of dependencies in a CNS model.

        Returns
        -------
        float | None
        """
        return self._num_dependencies

    @property
    def num_discrete_variables(self) -> float | None:
        """
        Number of discrete variables.

        Returns
        -------
        float | None
        """
        return self._num_discrete_variables

    @property
    def num_equations(self) -> float | None:
        """
        Number of equations.

        Returns
        -------
        float | None
        """
        return self._num_equations

    @property
    def num_infeasibilities(self) -> float | None:
        """
        Number of infeasibilities.

        Returns
        -------
        float | None
        """
        return self._num_infeasibilities

    @property
    def num_nonlinear_insts(self) -> float | None:
        """
        Number of nonlinear instructions.

        Returns
        -------
        float | None
        """
        return self._num_nonlinear_insts

    @property
    def num_nonlinear_zeros(self) -> float | None:
        """
        Number of nonlinear nonzeros.

        Returns
        -------
        float | None
        """
        return self._num_nonlinear_zeros

    @property
    def num_nonoptimalities(self) -> float | None:
        """
        Number of nonoptimalities.

        Returns
        -------
        float | None
        """
        return self._num_nonoptimalities

    @property
    def num_nonzeros(self) -> float | None:
        """
        Number of nonzero entries in the model coefficient matrix.

        Returns
        -------
        float | None
        """
        return self._num_nonzeros

    @property
    def num_mcp_redefinitions(self) -> float | None:
        """
        Number of MCP redefinitions.

        Returns
        -------
        float | None
        """
        return self._num_mcp_redefinitions

    @property
    def num_variables(self) -> float | None:
        """
        Number of variables.

        Returns
        -------
        float | None
        """
        return self._num_variables

    @property
    def num_bound_projections(self) -> float | None:
        """
        Number of bound projections during model generation.

        Returns
        -------
        float | None
        """
        return self._num_bound_projections

    @property
    def objective_estimation(self) -> float | None:
        """
        Estimate of the best possible solution for a mixed-integer model

        Returns
        -------
        float | None
        """
        return self._objective_estimation

    @property
    def objective_value(self) -> float | None:
        """
        Objective function value

        Returns
        -------
        float | None
        """
        return self._objective_value

    @property
    def used_model_type(self) -> float | None:
        """
        Integer number that indicates the used model type.

        Returns
        -------
        float | None
        """
        return self._used_model_type

    @property
    def model_generation_time(self) -> float | None:
        """
        Time GAMS took to generate the model in wall-clock seconds.

        Returns
        -------
        float | None
        """
        return self._model_generation_time

    @property
    def solve_model_time(self) -> float | None:
        """
        Time the solver used to solve the model in seconds

        Returns
        -------
        float | None
        """
        return self._solve_model_time

    @property
    def solve_status(self) -> SolveStatus | None:
        """
        Indicates the solver termination condition.

        Returns
        -------
        SolveStatus | None
        """
        return self._solve_status

    @property
    def sum_infeasibilities(self) -> float | None:
        """
        Sum of infeasibilities.

        Returns
        -------
        float | None
        """
        return self._sum_infeasibilities

    @property
    def solver_version(self) -> float | None:
        """
        Solver version.

        Returns
        -------
        float | None
        """
        return self._solver_version

    @property
    def external_module(self) -> str | None:
        """
        Name of the external module in which the external equations are implemented.
        By default, this parameter is set to None. When provided, it triggers the
        opening of the specified file using a File statement and incorporates the
        file into the model by adding it as an external module.

        This feature requires a solid understanding of programming, compilation,
        and linking processes. For more information, please refer to the
        https://gamspy.readthedocs.io/en/latest/user/advanced/external_equations.html .

        Returns
        -------
        str | None
        """
        return self._external_module

    @external_module.setter
    def external_module(self, value: str | None):
        if self._external_module_file is not None:
            self.container._add_statement(f"putclose {self._external_module_file};")

        write_model_statement = value != self._external_module
        self._external_module = None
        self._external_module_file = None

        if value is not None:
            filename = "f" + utils._get_unique_name()
            self._external_module_file = filename
            self._external_module = value
            self.container._add_statement(f"File {filename} / '{value}' /;")

        if write_model_statement:
            self.container._add_statement(self)

    def _generate_obj_var_and_equation(self) -> tuple[Variable, Equation]:
        variable = gp.Variable._constructor_bypass(
            self.container,
            f"{self.name}_objective_variable",
            domain=[],
        )
        equation = gp.Equation._constructor_bypass(
            self.container,
            f"{self.name}_objective",
            domain=[],
        )
        self._autogen_symbols.add(variable.name)
        self._autogen_symbols.add(equation.name)

        return variable, equation

    def _set_objective_variable(
        self,
        assignment: None | Variable | Operation | Expression | MathOp = None,
    ) -> Variable | None:
        """
        Returns objective variable. If the assignment is an Expression
        or an Operation (Sum, Product etc.), it automatically generates
        an objective variable and a equation.

        Returns
        -------
        Variable | None

        Raises
        ------
        TypeError
            In case assignment is not a Variable, Expression or an Operation.
        """

        if assignment is not None and not isinstance(
            assignment,
            (
                gp.Variable,
                expression.Expression,
                operation.Operation,
                gp.math.MathOp,
            ),
        ):
            raise TypeError(
                "Objective must be a Variable or an Expression but"
                f" {type(assignment)} given"
            )

        if self.sense == gp.Sense.FEASIBILITY:
            if assignment is not None:
                raise ValidationError(
                    "Cannot set an objective when the sense is FEASIBILITY!"
                )

            if self.problem in (Problem.CNS, Problem.MCP, Problem.EMP):
                return None

            variable, equation = self._generate_obj_var_and_equation()
            statement = expression.Expression(
                implicits.ImplicitEquation(
                    equation,
                    name=equation.name,
                    type=equation.type,
                    domain=[],
                ),
                "..",
                variable == 0,
            )
            self.container._add_statement(statement)
            equation._definition = statement
            equation.modified = False
            variable.modified = False

            if equation.name not in (symbol.name for symbol in self.equations):
                self.equations.append(equation)

            return variable

        if isinstance(
            assignment,
            (expression.Expression, operation.Operation, gp.math.MathOp),
        ):
            variable, equation = self._generate_obj_var_and_equation()

            # Sum((i,j),c[i,j]*x[i,j])->Sum((i,j),c[i,j]*x[i,j]) =e= var
            assignment = assignment == variable

            # equation .. Sum((i,j),c[i,j]*x[i,j]) =e= var
            statement = expression.Expression(
                implicits.ImplicitEquation(
                    equation,
                    name=equation.name,
                    type=equation.type,
                    domain=[],
                ),
                "..",
                assignment,
            )
            self.container._add_statement(statement)
            equation._definition = statement
            equation.modified = False
            variable.modified = False
            if equation.name not in (symbol.name for symbol in self.equations):
                self.equations.append(equation)

            return variable

        return assignment

    def _generate_solve_string(self) -> str:
        solve_statement = [f"solve {self.name} using {self.problem}"]

        if self.sense == gp.Sense.FEASIBILITY:
            # Set sense as min or max for feasibility
            self.sense = gp.Sense("MIN")

        if self.problem not in (Problem.MCP, Problem.CNS, Problem.EMP):
            solve_statement.append(str(self.sense))

        if self._objective_variable is not None:
            solve_statement.append(self._objective_variable.gamsRepr())

        return " ".join(solve_statement)

    def _add_runtime_options(self, options: Options, backend: str = "local") -> None:
        for key, value in options.model_dump(exclude_none=True).items():
            if key in MODEL_ATTR_OPTION_MAP:
                if isinstance(value, bool):
                    value = int(value)
                elif isinstance(value, str):
                    value = f"'{value}'"

                self.container._add_statement(
                    f"{self.name}.{MODEL_ATTR_OPTION_MAP[key]} = {value};\n"
                )
            elif key in EXECUTION_OPTIONS:
                if key == "loadpoint":
                    if isinstance(value, os.PathLike):
                        value = os.path.abspath(value)

                    if backend == "engine":
                        value = os.path.relpath(value, self.container.working_directory)

                self.container._add_statement(f"{EXECUTION_OPTIONS[key]} '{value}';\n")

    def _create_model_attributes(self) -> None:
        self.container._add_statement("$offListing")
        for attr_name in ATTRIBUTE_MAP:
            symbol_name = f"{self._generate_prefix}{attr_name}_{self._auto_id}"
            Symbol = gp.Parameter._constructor_bypass(self.container, symbol_name)
            Symbol.modified = False

            self.container._add_statement(f"{symbol_name} = {self.name}.{attr_name};")
        self.container._add_statement("$onListing")

    def _update_model_attributes(self) -> None:
        gdx_handle = utils._open_gdx_file(
            self.container.system_directory, self.container._gdx_out
        )

        for gams_attr, python_attr in ATTRIBUTE_MAP.items():
            symbol_name = f"{self._generate_prefix}{gams_attr}_{self._auto_id}"
            data = utils._get_scalar_data(
                self.container._gams2np, gdx_handle, symbol_name
            )

            if python_attr == "_status":
                setattr(self, python_attr, ModelStatus(data))
            elif python_attr == "_solve_status":
                status = SolveStatus(data)
                setattr(self, python_attr, status)

                if status in INTERRUPT_STATUS:
                    logger.warning(
                        f"The solve was interrupted! Solve status: {status.name}. "
                        "For further information, see https://gamspy.readthedocs.io/en/latest/reference/gamspy._model.html#gamspy.SolveStatus."
                    )
                elif status in ERROR_STATUS:
                    raise GamspyException(
                        f"Solve status: {status.name}. {ERROR_STATUS[status]}",
                        status.value,
                    )
            else:
                setattr(self, python_attr, data)

        utils._close_gdx_handle(gdx_handle)

    def computeInfeasibilities(self) -> dict[str, pd.DataFrame]:
        """
        Computes infeasabilities for all equations of the model

        Returns
        -------
        dict[str, pd.DataFrame]
            Dictionary of infeasibilities where equation names are keys and
            infeasibilities are values

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["i1", "i2"])
        >>> j = gp.Set(m, name="j", records=["j1", "j2", "j3"])
        >>> a = gp.Parameter(m, name="a", domain=i, records=[("i1", 350), ("i2", 600)])
        >>> b = gp.Parameter(m, name="b", domain=j, records=[("j1", 400), ("j2", 450), ("j3", 420)])
        >>> x = gp.Variable(m, name="x", domain=[i,j], type="Positive")
        >>> s = gp.Equation(m, name="s", domain=i)
        >>> d = gp.Equation(m, name="d", domain=j)
        >>> s[i] = gp.Sum(j, x[i, j]) <= a[i]
        >>> d[j] = gp.Sum(i, x[i, j]) >= b[j]
        >>> my_model = gp.Model(m, name="my_model", equations=m.getEquations(), problem="LP", sense="min", objective=gp.Sum((i, j), x[i, j]))
        >>> summary = my_model.solve()
        >>> infeasibilities = my_model.computeInfeasibilities()
        >>> infeasibilities["s"].infeasibility.item()
        320.0

        """
        infeas_dict = {}

        for equation in self.equations:
            if equation.records is None:
                continue

            infeas_rows = utils._calculate_infeasibilities(equation)
            infeas_dict[equation.name] = infeas_rows

            if equation._definition is not None:
                names = equation._definition._find_all_symbols()
                names = [
                    name
                    for name in names
                    if name in self.container.data
                    and isinstance(self.container[name], gp.Variable)
                ]

                for name in names:
                    variable = self.container[name]
                    infeas_rows = utils._calculate_infeasibilities(variable)
                    infeas_dict[variable.name] = infeas_rows

        return infeas_dict

    def convert(
        self,
        path: str | Path,
        file_format: FileFormat | Sequence[FileFormat],
        options: ConvertOptions | None = None,
    ) -> None:
        """
        Converts the model to one or more specified file formats.

        Parameters
        ----------
        path : str | Path
            Path to the directory where the converted model files will be saved.
        file_format : FileFormat | Sequence[FileFormat]
            File format(s) to convert the model to. Can be a single FileFormat or a list of FileFormats.
        options : ConvertOptions, optional
            Additional options to customize the conversion process.

        Raises
        ------
        ValueError
            If the specified file format is not supported.

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> v = gp.Variable(m, "v")
        >>> e = gp.Equation(m, "e", definition= v == 5)
        >>> my_model = gp.Model(m, "my_model", problem="LP", equations=[e])
        >>> my_model.convert("tmp", gp.FileFormat.GAMS)
        >>> my_model.convert("tmp", [gp.FileFormat.GAMS, gp.FileFormat.AMPL])

        """
        path = Path(path)
        os.makedirs(path, exist_ok=True)
        solver_options = get_convert_solver_options(path, file_format, options)
        self.solve(solver="convert", solver_options=solver_options)

    def getEquationListing(
        self,
        infeasibility_threshold: float | None = None,
    ) -> str:
        """
        Returns the generated equations.

        Parameters
        ----------
        infeasibility_threshold: float, optional
            Filters out equations with infeasibilities that are above this value.

        Returns
        -------
        str
        """
        listings = []
        for equation in self.equations:
            listing = equation.getEquationListing(
                infeasibility_threshold=infeasibility_threshold
            )

            if listing:
                listings.append(listing)

        return "\n".join(listings)

    def getVariableListing(self) -> str:
        """
        Returns the variable listing.

        Returns
        -------
        str
        """
        if not hasattr(self, "_variables"):
            raise ValidationError(
                "The model must be solved with `variable_listing_limit` option for this functionality to work."
            )

        listings = []
        for variable in self._variables:
            listings.append(variable.getVariableListing())

        return "\n".join(listings)

    def interrupt(self) -> None:
        """Sends interrupt signal to the running job."""
        self.container._interrupt()

    def freeze(
        self,
        modifiables: list[Parameter | ImplicitParameter],
        options: Options | None = None,
    ) -> None:
        """
        Instantiates a model instance. After calling freeze, only modifiables can be modified.

        Parameters
        ----------
        modifiables : list[Parameter  |  ImplicitParameter]
            Modifiable symbols.
        options : Options | None, optional
            GAMSPy options, by default None
        """
        self._is_frozen = True
        if options is None:
            options = self.container._options

        self.instance = ModelInstance(
            self.container, self, modifiables, options, self.container.output
        )

    def unfreeze(self) -> None:
        """Unfreezes the model"""
        self._is_frozen = False
        self.instance.close_license_session()

    def solve(
        self,
        solver: str | None = None,
        options: Options | None = None,
        solver_options: dict | str | Path | None = None,
        freeze_options: FreezeOptions | None = None,
        output: TextIO | None = None,
        backend: Literal["local", "engine", "neos"] = "local",
        client: EngineClient | NeosClient | None = None,
        load_symbols: list[Symbol] | None = None,
    ) -> pd.DataFrame | None:
        """
        Solves the model with given options.

        Parameters
        ----------
        solver : str, optional
            Solver name
        options : Options, optional
            GAMSPy options.
        solver_options : dict | str | Path, optional
            Dictionary of solver options or path to an existing option file.
        freeze_options : FreezeOptions, optional
            Options to solve a frozen model.
        output : TextIO, optional
            Output redirection target. Set this to sys.stdout to see the output in your terminal.
        backend : str, optional
            Backend to run on
        client : EngineClient, NeosClient, optional
            EngineClient to communicate with GAMS Engine or NEOS Client to communicate with NEOS Server
        load_symbols : list[Symbol], optional
            Specifies the symbols that need to be loaded. If not given, all symbols are loaded after solve.

        Returns
        -------
        DataFrame, optional
            Summary of the solve

        Raises
        ------
        ValidationError
            In case engine_config is not provided for `engine` backend or
            neos_client is not provided for `neos` backend.
        ValueError
            In case problem is not in possible problem types
        ValueError
            In case sense is different than "MIN" or "MAX"

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> v = gp.Variable(m, "v")
        >>> e = gp.Equation(m, "e", definition= v == 5)
        >>> my_model = gp.Model(m, "my_model", problem="LP", equations=[e], sense="max", objective=v)
        >>> solved = my_model.solve()

        """
        if output is None:
            output = self.container.output

        if solver is None:
            solver = utils.getDefaultSolvers(self.container.system_directory)[
                str(self.problem).upper()
            ].lower()
        else:
            if not isinstance(solver, str):
                raise TypeError(
                    f"`solver` argument must be of type `str` but given `{type(solver)}`"
                )
            solver = solver.lower()

        if solver == "conopt":
            solver = "conopt4"

        validation.validate_solver_args(
            self.container.system_directory,
            backend,
            solver,
            self.problem,
            options,
            output,
            load_symbols,
        )
        validation.validate_equations(self)

        if options is None:
            options = self.container._options

        if isinstance(solver_options, (str, Path)):
            solver_options = Path(solver_options)

        # Only for local until GAMS Engine and NEOS Server backends adopt the new GP_SolveLine option.
        if backend == "local":
            frame = inspect.currentframe().f_back
            assert isinstance(options, gp.Options)
            options._frame = frame

        if solver_options is not None:
            self.container.writeSolverOptions(solver, solver_options)

        if self._is_frozen:
            instance_options = FreezeOptions()

            if freeze_options is not None:
                instance_options = freeze_options

            summary = self.instance.solve(
                solver, instance_options, solver_options, output
            )
            return summary

        self.container._add_statement(self.getDeclaration())
        self._add_runtime_options(options, backend)
        self.container._add_statement(self._generate_solve_string() + "\n")
        self._create_model_attributes()
        options._set_model_info(solver, self.problem, solver_options)

        runner = backend_factory(
            self.container,
            options,
            solver,
            solver_options,
            output,
            backend,
            client,
            self,
            load_symbols,
        )

        summary = runner.run(relaxed_domain_mapping=False, gams_to_gamspy=True)

        if IS_MIRO_INIT:
            miro._write_default_gdx_miro(self.container)

        return summary

    def getDeclaration(self) -> str:
        """
        Declaration of the Model in GAMS

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> v = gp.Variable(m, "v")
        >>> e = gp.Equation(m, "e", definition= v == 5)
        >>> my_model = gp.Model(m, "my_model", problem="LP", equations=[e])
        >>> my_model.getDeclaration()
        'Model my_model / e,my_model_objective /;'

        """

        def build_str_with_new_lines(elements: list[str]) -> str:
            num_elements = len(elements)
            if num_elements < MAX_NUM_MODEL_ELEMS:
                return ",".join(elements)

            count = 0
            return_str = ""
            while count < num_elements:
                batch = elements[count : count + MAX_NUM_MODEL_ELEMS]
                return_str += ",".join(batch) + "\n"
                count += MAX_NUM_MODEL_ELEMS

            return return_str

        equations_in_matches = []
        if self._matches is not None:
            for key in self._matches:
                if isinstance(key, gp.Equation):
                    equations_in_matches.append(key)
                else:
                    equations_in_matches += list(key)

        equations = []
        for equation in self.equations:
            if self._matches is not None:
                if equation not in equations_in_matches:
                    equations.append(equation.name)
            else:
                equations.append(equation.name)

        equations_str = build_str_with_new_lines(equations)

        if self._matches is not None:
            matches = []
            for key, value in self._matches.items():
                if isinstance(key, gp.Equation):
                    if isinstance(value, gp.Variable):
                        matches.append(f"{key.name}.{value.name}")
                    else:
                        matches.append(
                            f"{key.name}:({'|'.join([variable.name for variable in value])})"
                        )
                else:
                    assert isinstance(value, gp.Variable)
                    matches.append(
                        f"({'|'.join([equation.name for equation in key])}):{value.name}"
                    )

            matches_str = build_str_with_new_lines(matches)

            equations_str = (
                ",".join([equations_str, matches_str]) if equations else matches_str
            )

        if self._external_module_file is not None:
            equations_str = ",".join([equations_str, self._external_module_file])

        if self._limited_variables is not None:
            limited_variables_str = ",".join(
                [variable.gamsRepr() for variable in self._limited_variables]
            )
            equations_str += "," + limited_variables_str

        model_str = f"Model {self.name}"
        if self.description:
            model_str += ' "' + self.description + '"'

        if equations_str != "":
            model_str += f" / {equations_str} /"
        model_str += ";"

        return model_str

    def toGams(
        self,
        path: str | Path,
        options: Options | None = None,
        *,
        dump_gams_state: bool = False,
    ) -> None:
        """
        Generates GAMS model under path/<model_name>.gms.

        Parameters
        ----------
        path : str | Path
            Path to the directory which will contain the GAMS model.
        options : Options | None, optional
            GAMSPy options, by default None
        dump_gams_state : bool, optional
            Whether to dump the state as a GAMS save file, by default False

        Raises
        ------
        ValidationError
            In case the given options is not of type gp.Options.

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> v = gp.Variable(m, "v")
        >>> e = gp.Equation(m, "e", definition= v == 5)
        >>> my_model = gp.Model(m, "my_model", problem="LP", equations=[e])
        >>> my_model.toGams("tmp")  # doctest: +ELLIPSIS
        ================================================================================
        GAMS (.gms) file has been generated under ...
        ================================================================================

        """
        if options is not None and not isinstance(options, gp.Options):
            raise ValidationError(
                f"`options` must be of type gp.Options of found {type(options)}"
            )

        path = Path(path)
        converter = GamsConverter(self, path, options, dump_gams_state)
        converter.convert()

    def toLatex(
        self,
        path: str | Path,
        rename: dict[str, str] | None = None,
        *,
        generate_pdf: bool = False,
    ) -> None:
        """
        Generates a latex file that contains the model definition under path/<model_name>.tex

        Parameters
        ----------
        path : str | Path
            Path to the directory which will contain the .tex file.
        rename: dict[str, str], optional
            A dictionary to rename symbols in the LaTeX file. Keys are GAMSPy symbol names
            and values are the names that will be used in the LaTeX file.
        generate_pdf: bool, False by default
            Generates a pdf file if it is set. Requires pdflatex to be installed.

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> v = gp.Variable(m, "v")
        >>> e = gp.Equation(m, "e", definition= v == 5)
        >>> my_model = gp.Model(m, "my_model", problem="LP", equations=[e])
        >>> my_model.toLatex("tmp")  # doctest: +ELLIPSIS
        ================================================================================
        LaTeX (.tex) file has been generated under ...
        ================================================================================

        """
        path = Path(path)
        converter = LatexConverter(self, path, rename)
        converter.convert()

        if generate_pdf:
            converter.to_pdf()



from __future__ import annotations

import importlib
import itertools
import uuid
from typing import TYPE_CHECKING, Any

import gamspy as gp
import gamspy.formulations.utils as utils
from gamspy.exceptions import ValidationError
from gamspy.formulations.ml.decision_tree_struct import DecisionTreeStruct
from gamspy.formulations.ml.regression_tree import RegressionTree

if TYPE_CHECKING:
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.tree import DecisionTreeRegressor

    from gamspy import (
        Alias,
        Equation,
        Parameter,
        Set,
        Variable,
    )


class RandomForest:
    """
    Formulation generator for Random Forests in GAMSPy.

    Parameters
    ----------
    container : Container
        Container that will contain the new variable and equations.
    ensemble: RandomForestRegressor | None
        - A fitted `sklearn.ensemble.RandomForestRegressor` instance,
        - If `sklearn.ensemble.RandomForestRegressor` is not utilized, the ensembled trees information
          can be supplied via a list of `DecisionTreeStruct` dataclasse instances, which represents
          the same components as those in `sklearn.tree`.
          See :meth:`DecisionTreeStruct <gamspy.formulations.DecisionTreeStruct>` for details on required attributes.

    name_prefix : str | None
        Prefix for generated GAMSPy symbols, by default None which means
        random prefix. Using the same name_prefix in different formulations causes name
        conflicts. Do not use the same name_prefix again.

    Examples
    --------
    >>> import gamspy as gp
    >>> import numpy as np
    >>> from gamspy.math import dim
    >>> np.random.seed(42)
    >>> m = gp.Container()
    >>> in_data = np.random.randint(0, 10, size=(5, 2))
    >>> out_data = np.random.randint(1, 3, size=(5, 1))
    >>> tree1_attribute = {
    ...     "capacity": 7,
    ...     "children_left": np.array([ 1, -1,  3, -1,  5, -1, -1]),
    ...     "children_right": np.array([ 2, -1,  4, -1,  6, -1, -1]),
    ...     "feature": np.array([ 1, -2,  0, -2,  1, -2, -2]),
    ...     "n_features": 2,
    ...     "threshold": np.array([ 2. , -2. ,  5.5, -2. ,  8.5, -2. , -2. ]),
    ...     "value": np.array([[1.6 ],[1.  ],[1.75],[2.  ],[1.5 ],[1.  ],[2.  ]])
    ... }
    >>> tree2_attribute = {
    ...     "capacity": 3,
    ...     "children_left": np.array([ 1, -1, -1]),
    ...     "children_right": np.array([ 2, -1, -1]),
    ...     "feature": np.array([ 0, -2, -2]),
    ...     "n_features": 2,
    ...     "threshold": np.array([ 1.5, -2. , -2. ]),
    ...     "value": np.array([[1.4],[1. ],[2. ]])
    ... }
    >>> forest = [gp.formulations.DecisionTreeStruct(**tree1_attribute), gp.formulations.DecisionTreeStruct(**tree2_attribute)]
    >>> dt_model = gp.formulations.RandomForest(m, forest)
    >>> x = gp.Variable(m, "x", domain=dim((5, 2)), type="positive")
    >>> x.up[:, :] = 10
    >>> y, eqns = dt_model(x)
    >>> set_of_samples = y.domain[0]
    >>> set_of_samples.name
    'DenseDim5_1'

    """

    def __init__(
        self,
        container: gp.Container,
        ensemble: RandomForestRegressor | list[DecisionTreeStruct],
        name_prefix: str | None = None,
    ):
        if not isinstance(container, gp.Container):
            raise ValidationError(f"{container} is not a gp.Container.")

        def _validate_ensemble(
            ensemble,
        ) -> list[DecisionTreeRegressor | DecisionTreeStruct]:
            if isinstance(ensemble, list) and all(
                isinstance(item, DecisionTreeStruct) for item in ensemble
            ):
                return ensemble
            else:
                try:
                    sklearn_forest = importlib.import_module("sklearn.ensemble")
                    if isinstance(ensemble, sklearn_forest.RandomForestRegressor):
                        if not hasattr(ensemble, "estimators_"):
                            raise ValidationError(
                                f"{ensemble} must be a trained/fitted instance of >sklearn.ensemble.RandomForestRegressor<."
                            )
                        return ensemble.estimators_
                    else:
                        raise ValidationError(
                            f"{ensemble} must be an instance of either >sklearn.ensemble.RandomForestRegressor< or a list of >DecisionTreeStruct<"
                        )
                except ModuleNotFoundError:
                    raise ValidationError(
                        ">sklearn.ensemble< module not found."
                    ) from None

        self.container = container

        if name_prefix is None:
            name_prefix = str(uuid.uuid4()).split("-")[0]

        self._name_prefix = name_prefix

        self._list_of_trees: list[RegressionTree] = []

        for tree in _validate_ensemble(ensemble):
            rt_instance = RegressionTree(
                self.container,
                regressor=tree,
                name_prefix=self._name_prefix,
            )
            self._list_of_trees.append(rt_instance)

    def __call__(
        self,
        input: gp.Parameter | gp.Variable,
        M: float | None = None,
    ) -> tuple[gp.Variable, list[gp.Equation]]:
        rf_out_list: list[gp.Variable] = []
        rf_eqn_list: list[gp.Equation] = []
        previous_value = gp.get_option("DOMAIN_VALIDATION")
        gp.set_options({"DOMAIN_VALIDATION": 0})

        set_records_total: dict[Set | Alias | Parameter | Variable | Equation, Any] = {}

        results = (
            regression_tree._yield_call(input, M)
            for regression_tree in self._list_of_trees
        )

        zipped_results = zip(*results, strict=False)
        dt_outs = next(zipped_results)
        rf_out_list.extend(dt_outs)

        set_records_iter = next(zipped_results)
        set_of_output_dim = None
        for item, set_records_dict in set_records_iter:
            set_records_total.update(set_records_dict)
            set_of_output_dim = item

        self.container.setRecords(set_records_total)

        rf_eqn_list = list(itertools.chain.from_iterable(next(zipped_results)))

        set_of_samples = input.domain[0]

        out = gp.Variable._constructor_bypass(
            self.container,
            name=utils._generate_name("v", self._name_prefix, "real_output"),
            domain=[set_of_samples, set_of_output_dim],  # type: ignore
        )

        rf_eqn = gp.Equation._constructor_bypass(
            self.container,
            name=utils._generate_name("e", self._name_prefix, "rf_eqn"),
            domain=[set_of_samples, set_of_output_dim],  # type: ignore
            description="predicted out times number of estimators should be equal to the random forest out",
        )

        self.container._synch_with_gams(gams_to_gamspy=True)
        rf_eqn[...] = len(self._list_of_trees) * out == sum(rf_out_list)
        rf_eqn_list.append(rf_eqn)
        gp.set_options({"DOMAIN_VALIDATION": previous_value})

        return out, rf_eqn_list



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

import gamspy._algebra.expression as expression
import gamspy._algebra.operable as operable
import gamspy._algebra.operation as operation
import gamspy._symbols as syms
import gamspy._symbols.implicits as implicits
import gamspy._validation as validation
import gamspy.utils as utils
from gamspy._symbols.implicits.implicit_symbol import ImplicitSymbol
from gamspy._symbols.implicits.implicit_variable import ImplicitVariable
from gamspy.exceptions import ValidationError
from gamspy.math.matrix import permute

if TYPE_CHECKING:
    from collections.abc import Sequence
    from types import EllipsisType

    import pandas as pd

    from gamspy import (
        Alias,
        Equation,
        Parameter,
        Set,
        Variable,
    )
    from gamspy._algebra.expression import Expression
    from gamspy._algebra.operation import Operation

logger = logging.getLogger("GAMSPy")
logger.setLevel(logging.WARNING)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.WARNING)
formatter = logging.Formatter("[%(name)s - %(levelname)s] %(message)s")
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)

ATTR_MAPPING = {
    "l": "level",
    "m": "marginal",
    "lo": "lower",
    "up": "upper",
    "scale": "scale",
    "range": "range",
    "slack": "slack",
    "slacklo": "slacklo",
    "slackup": "slackup",
    "infeas": "infeas",
}

SET_ATTR_MAPPING = {
    "pos": "position",
    "ord": "order",
    "off": "off",
    "rev": "reverse",
    "uel": "uel_position",
    "len": "length",
    "tlen": "text_length",
    "val": "value",
    "tval": "text_value",
    "first": "is_first",
    "last": "is_last",
}


class ImplicitParameter(ImplicitSymbol, operable.Operable):
    def __init__(
        self,
        parent: Set | Alias | Parameter | Variable | Equation,
        name: str,
        domain: list[Set | str] = [],
        records: Any | None = None,
        permutation: list[int] | None = None,
        scalar_domains: list[tuple[int, Set]] | None = None,
    ) -> None:
        """Implicit Parameter

        Parameters
        ----------
        parent : Parameter | Variable | Equation
        name : str
        domain : list[Set | str], optional
        records : Any, optional
        """
        super().__init__(parent, name, domain, permutation, scalar_domains)
        self._records = records
        self._assignment = None

    def __getitem__(
        self,
        indices: EllipsisType | slice | Sequence | str | implicits.ImplicitSet,
    ) -> ImplicitParameter:
        domain = validation.validate_domain(self, indices)

        return ImplicitParameter(
            parent=self.parent,
            name=self.name,
            domain=domain,
            permutation=self.permutation,
            scalar_domains=self._scalar_domains,
        )

    def __setitem__(
        self,
        indices: EllipsisType | slice | Sequence | str | implicits.ImplicitSet,
        rhs: Expression | Operation | ImplicitParameter | int | float,
    ) -> None:
        if (
            isinstance(self.parent, (syms.Variable, syms.Equation))
            and len(self.parent.domain) > 0
            and all(len(elem) == 0 for elem in self.parent.domain)
        ):
            logger.warning(
                f"Domain was not initialized. Default values for {self.gamsRepr()} will be used."
            )
        # self[domain] = rhs
        domain = validation.validate_domain(self, indices)

        if isinstance(rhs, float):
            rhs = utils._map_special_values(rhs)  # type: ignore

        statement = expression.Expression(
            ImplicitParameter(
                parent=self.parent,
                name=self.name,
                domain=domain,
                permutation=self.permutation,
                scalar_domains=self._scalar_domains,
            ),
            "=",
            rhs,
        )

        statement._validate_definition(utils._unpack(domain))

        self.container._add_statement(statement)
        self.parent._assignment = statement

        self.container._synch_with_gams(gams_to_gamspy=True)
        self.parent._winner = "gams"

    def __eq__(self, other):
        op = "eq"
        if isinstance(
            other,
            (ImplicitVariable, expression.Expression, operation.Operation),
        ):
            op = "=e="
        return expression.Expression(self, op, other)

    def __ne__(self, other):
        return expression.Expression(self, "ne", other)

    def __repr__(self) -> str:
        return f"ImplicitParameter(parent={self.parent}, name='{self.name}', domain={self.domain}, permutation={self.permutation}), parent_scalar_domains={self.parent_scalar_domains})"

    @property
    def records(self) -> pd.DataFrame | None:
        if self.parent.records is None:
            return None

        if isinstance(self.parent, (syms.Set, syms.Alias)):
            temp_name = "p" + utils._get_unique_name()
            temp_param = syms.Parameter._constructor_bypass(
                self.container, temp_name, [self.parent, "*"]
            )
            column_name = SET_ATTR_MAPPING[self.name.split(".")[1]]
            temp_param[self.parent, column_name] = self
            del self.container.data[temp_name]
            return temp_param.records
        elif isinstance(self.parent, syms.Parameter):
            temp_name = "p" + utils._get_unique_name()
            temp_param = syms.Parameter._constructor_bypass(
                self.container, temp_name, self.parent.domain
            )
            domain = list(self.domain)
            for i, d in self._scalar_domains:
                domain.insert(i, d)

            if domain == []:
                domain = [...]

            temp_param[domain] = self[...]
            del self.container.data[temp_name]

            recs = temp_param.records
            return recs
        elif isinstance(self.parent, (syms.Variable, syms.Equation)):
            extension = self.name.split(".")[-1]
            temp_name = "ip" + utils._get_unique_name()
            temp_param = syms.Parameter._constructor_bypass(
                self.container, temp_name, self.parent.domain
            )
            domain = list(self.domain)
            for i, d in self._scalar_domains:
                domain.insert(i, d)

            if domain == []:
                domain = [...]

            temp_param[domain] = self
            del self.container.data[temp_name]

            recs = temp_param.records
            if recs is None:
                return recs

            extension = ATTR_MAPPING[extension]
            columns = recs.columns.to_list()
            columns[columns.index("value")] = extension
            recs.columns = columns

            return recs

        return None

    @property
    def T(self) -> ImplicitParameter:
        """See gamspy.ImplicitParameter.t"""
        return self.t()

    def t(self) -> ImplicitParameter:
        """Returns an ImplicitParameter derived from this
        implicit parameter by swapping its last two indices.
        This operation does not generate a new parameter in GAMS.

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=['i1','i2'])
        >>> j = gp.Set(m, "j", records=['j1','j2'])
        >>> v = gp.Parameter(m, "v", domain=[i, j])
        >>> v_t = v.t() # v_t is an ImplicitParameter
        >>> v_t_t = v_t.t() # you can get transpose of ImplicitParameter as well
        >>> v_t_t.domain
        [Set(name='i', domain=['*']), Set(name='j', domain=['*'])]

        """
        dims = list(range(len(self.domain)))
        if len(dims) < 2:
            raise ValidationError(
                "Parameter must contain at least 2 dimensions to transpose"
            )

        x = dims[-1]
        dims[-1] = dims[-2]
        dims[-2] = x
        return permute(self, dims)  # type: ignore

    def gamsRepr(self) -> str:
        """Representation of the parameter in GAMS syntax.

        Returns:
            str: String representation of the parameter in GAMS syntax.
        """
        representation = self.name
        domain = list(self.domain)
        if domain and self.permutation is not None:
            domain = utils._permute_domain(domain, self.permutation)

        for i, d in self._scalar_domains:
            domain.insert(i, d)

        if domain:
            representation += utils._get_domain_str(domain)

        return representation



from __future__ import annotations

from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import rich
import rich.table
import typer

import gamspy.utils as utils
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from collections.abc import Callable

app = typer.Typer(
    rich_markup_mode="rich",
    short_help="To show your license and gamspy_base directory.",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy show license | gamspy show base",
    context_settings={"help_option_names": ["-h", "--help"]},
)

LICENSE_TYPE_MAP = {
    "00": "Demo license. All solvers available with size limitations.\nSee https://www.gams.com/latest/docs/UG_License.html#UG_License_Additional_Solver_Limits",
    "05": "Community license. All solvers available with size limitations.\nSee https://www.gams.com/latest/docs/UG_License.html#UG_License_Additional_Solver_Limits",
    "07": "GAMSPy license",
    "08": "GAMSPy++ license",
    "09": "GAMS & GAMSPy++ license",
}

SOLVER_PAGE_MAP = {
    "ANTIGONE": "https://www.gams.com/latest/docs/S_ANTIGONE.html",
    "BARON": "https://www.gams.com/latest/docs/S_BARON.html",
    "CBC": "https://www.gams.com/latest/docs/S_CBC.html",
    "CONOPT": "https://www.gams.com/latest/docs/S_CONOPT4.html",
    "CONVERT": "https://www.gams.com/latest/docs/S_CONVERT.html",
    "COPT": "https://www.gams.com/latest/docs/S_COPT.html",
    "CPLEX": "https://www.gams.com/latest/docs/S_CPLEX.html",
    "DE": "https://www.gams.com/latest/docs/S_DE.html",
    "DECISC": "https://www.gams.com/latest/docs/S_DECIS.html",
    "DECISM": "https://www.gams.com/latest/docs/S_DECIS.html",
    "DICOPT": "https://www.gams.com/latest/docs/S_DICOPT.html",
    "EXAMINER": "https://www.gams.com/latest/docs/S_EXAMINER.html",
    "GAMSCHK": "https://www.gams.com/latest/docs/S_GAMSCHK.html",
    "GUROBI": "https://www.gams.com/latest/docs/S_GUROBI.html",
    "HIGHS": "https://www.gams.com/latest/docs/S_HIGHS.html",
    "IPOPT": "https://www.gams.com/latest/docs/S_IPOPT.html",
    "IPOPTH": "https://www.gams.com/latest/docs/S_IPOPT.html",
    "JAMS": "https://www.gams.com/latest/docs/S_JAMS.html",
    "KESTREL": "https://www.gams.com/latest/docs/S_KESTREL.html",
    "KNITRO": "https://www.gams.com/latest/docs/S_KNITRO.html",
    "LINDO": "https://www.gams.com/latest/docs/S_LINDO.html",
    "LINDOGLOBAL": "https://www.gams.com/latest/docs/S_LINDO.html",
    "LOGMIP": "https://www.gams.com/latest/docs/S_JAMS.html",
    "MILES": "https://www.gams.com/latest/docs/S_MILES.html",
    "MINOS": "https://www.gams.com/latest/docs/S_MINOS.html",
    "MOSEK": "https://www.gams.com/latest/docs/S_MOSEK.html",
    "MPSGE": "https://www.gams.com/50/docs/UG_MPSGE_Intro.html",
    "NLPEC": "https://www.gams.com/latest/docs/S_NLPEC.html",
    "ODHCPLEX": "https://www.gams.com/latest/docs/S_ODHCPLEX.html",
    "PATH": "https://www.gams.com/latest/docs/S_PATH.html",
    "QUADMINOS": "https://www.gams.com/latest/docs/S_MINOS.html",
    "RESHOP": "https://www.gams.com/latest/docs/S_RESHOP.html",
    "SBB": "https://www.gams.com/latest/docs/S_SBB.html",
    "SCIP": "https://www.gams.com/latest/docs/S_SCIP.html",
    "SHOT": "https://www.gams.com/latest/docs/S_SHOT.html",
    "SNOPT": "https://www.gams.com/latest/docs/S_SNOPT.html",
    "SOPLEX": "https://www.gams.com/latest/docs/S_SOPLEX.html",
    "XPRESS": "https://www.gams.com/latest/docs/S_XPRESS.html",
}

# cm: component_map, ia: is_academic
EVALUATIONS: dict[str, Callable[[dict[str, bool], bool], bool]] = {
    "ANTIGONE": lambda cm, ia: cm["AT"]
    and (cm["CP"] or cm["CL"])
    and (cm["CO"] or cm["SN"]),
    "BARON": lambda cm, ia: cm["BA"],
    "CBC": lambda cm, ia: True,
    "CONOPT": lambda cm, ia: cm["CO"],
    "CONVERT": lambda cm, ia: True,
    "COPT": lambda cm, ia: cm["CT"] or cm["CK"],
    "CPLEX": lambda cm, ia: cm["CP"] or cm["CL"],
    "DE": lambda cm, ia: True,
    "DECISC": lambda cm, ia: cm["DE"] and (cm["CP"] or cm["CL"]),
    "DECISM": lambda cm, ia: cm["DE"] and cm["M5"],
    "DICOPT": lambda cm, ia: cm["DI"],
    "EXAMINER": lambda cm, ia: True,
    "GAMSCHK": lambda cm, ia: True,
    "GUROBI": lambda cm, ia: cm["GU"] or cm["GL"],
    "HIGHS": lambda cm, ia: cm["HI"] or ia,
    "IPOPT": lambda cm, ia: True,
    "IPOPTH": lambda cm, ia: cm["IP"] or ia,
    "JAMS": lambda cm, ia: True,
    "KESTREL": lambda cm, ia: True,
    "KNITRO": lambda cm, ia: cm["KN"],
    "LINDO": lambda cm, ia: cm["LD"],
    "LINDOGLOBAL": lambda cm, ia: cm["LD"] or cm["LI"],
    "LOGMIP": lambda cm, ia: True,
    "MILES": lambda cm, ia: True,
    "MINOS": lambda cm, ia: cm["M5"],
    "MOSEK": lambda cm, ia: cm["MB"] or cm["ML"],
    "MPSGE": lambda cm, ia: cm["GE"],
    "NLPEC": lambda cm, ia: True,
    "ODHCPLEX": lambda cm, ia: (ia or cm["OD"]) and (cm["CP"] or cm["CL"]),
    "PATH": lambda cm, ia: cm["PT"],
    "QUADMINOS": lambda cm, ia: cm["M5"],
    "RESHOP": lambda cm, ia: True,
    "SBB": lambda cm, ia: cm["SB"],
    "SCIP": lambda cm, ia: ia or cm["SC"],
    "SHOT": lambda cm, ia: True,
    "SNOPT": lambda cm, ia: cm["SN"],
    "SOPLEX": lambda cm, ia: ia or cm["SC"],
    "XPRESS": lambda cm, ia: cm["XP"] or cm["XL"] or cm["XS"] or cm["XX"] or cm["XG"],
}

LICENSE_COMPONENTS = {
    "AT": "ANTIGONE                 https://www.gams.com/latest/docs/S_ANTIGONE.html",
    "BA": "BARON                    https://www.gams.com/latest/docs/S_BARON.html",
    "CL": "CPLEX (Link only)        https://www.gams.com/latest/docs/S_CPLEX.html",
    "CO": "CONOPT                   https://www.gams.com/latest/docs/S_CONOPT4.html",
    "CP": "CPLEX                    https://www.gams.com/latest/docs/S_CPLEX.html",
    "CT": "COPT                     https://www.gams.com/latest/docs/S_COPT.html",
    "CK": "COPT (Link only)         https://www.gams.com/latest/docs/S_COPT.html",
    "DE": "DECIS                    https://www.gams.com/latest/docs/S_DECIS.html",
    "DI": "DICOPT                   https://www.gams.com/latest/docs/S_DICOPT.html",
    "EC": "ALPHAECP                 https://www.gams.com/latest/docs/S_ALPHAECP.html",
    "GE": "MPSGE                    https://www.gams.com/latest/docs/UG_MPSGE.html",
    "GL": "GUROBI (Link only)       https://www.gams.com/latest/docs/S_GUROBI.html",
    "GU": "GUROBI                   https://www.gams.com/latest/docs/S_GUROBI.html",
    "HI": "HIGHS                    https://www.gams.com/latest/docs/S_HIGHS.html",
    "IP": "IPOPT/IPOPTH             https://www.gams.com/latest/docs/S_IPOPT.html",
    "KN": "KNITRO                   https://www.gams.com/latest/docs/S_KNITRO.html",
    "LD": "LINDO                    https://www.gams.com/latest/docs/S_LINDO.html",
    "LI": "LINDOGlobal              https://www.gams.com/latest/docs/S_LINDO.html",
    "M5": "MINOS/QUADMINOS          https://www.gams.com/latest/docs/S_MINOS.html",
    "MB": "MOSEK                    https://www.gams.com/latest/docs/S_MOSEK.html",
    "ML": "MOSEK (Link only)        https://www.gams.com/latest/docs/S_MOSEK.html",
    "OD": "ODHCPLEX                 https://www.gams.com/latest/docs/S_ODHCPLEX.html",
    "PT": "PATH/PATHNLP             https://www.gams.com/latest/docs/S_PATH.html",
    "SB": "SBB                      https://www.gams.com/latest/docs/S_SBB.html",
    "SC": "SCIP/SOPLEX              https://www.gams.com/latest/docs/S_SCIP.html",
    "SN": "SNOPT                    https://www.gams.com/latest/docs/S_SNOPT.html",
    "XL": "XPRESS (Link only)       https://www.gams.com/latest/docs/S_XPRESS.html",
    "XP": "XPRESS LP/MIP            https://www.gams.com/latest/docs/S_XPRESS.html",
    "XS": "XPRESS LP/SLP            https://www.gams.com/latest/docs/S_XPRESS.html",
    "XX": "XPRESS LP/SLP/MIP        https://www.gams.com/latest/docs/S_XPRESS.html",
    "XG": "XPRESS LP/SLP/MIP/Global https://www.gams.com/latest/docs/S_XPRESS.html",
}


def print_expiration_date(lines: list[str]) -> None:
    month = lines[3][0:2]
    m = month[1] if month[0] == "0" else month[0]
    mval = 0
    if "1" <= m <= "9":
        mval = ord(m) - ord("0")
    elif "A" <= m <= "Z":
        mval = ord(m) - ord("A") + 10
    elif "a" <= m <= "z":
        mval = ord(m) - ord("a") + 36

    assert mval > 0
    expiration = datetime.strptime(lines[0][48:54], "%y%m%d") + timedelta(
        days=mval * 30
    )
    if month[0] == "0":
        rich.print(
            f"\n[bold]License expiration date[/bold]: {expiration.strftime('%Y-%m-%d')}"
        )
    else:
        rich.print(
            f"\n[bold]License M&S expiration date[/bold]: {expiration.strftime('%Y-%m-%d')}"
        )


def print_licensed_solvers(lines: list[str], verbose: bool) -> None:
    component_map: dict[str, bool] = dict.fromkeys(LICENSE_COMPONENTS.keys(), False)

    is_academic = lines[0][59] == "A"
    license_number = lines[2][0:2]
    rich.print(f"[bold]License type[/bold]: {LICENSE_TYPE_MAP[license_number]}")

    if license_number not in ["00", "05"]:
        licensed_components = []
        for i in range(2, len(lines[2]), 2):
            c = lines[2][i : i + 2]
            if c == "__":
                break
            component_map[c] = True
            licensed_components.append(c)

        if verbose and licensed_components:
            rich.print("\n[bold]Licensed components:[/bold]")
            for c in licensed_components:
                print(f"  {LICENSE_COMPONENTS[c]}")

        licensed_solvers = []
        for solver, is_licensed_func in EVALUATIONS.items():
            if is_licensed_func(component_map, is_academic):
                licensed_solvers.append(solver)

        if licensed_solvers:
            print()
            table = rich.table.Table(title="Licensed Solvers")
            table.add_column("Solver Name", style="green")
            table.add_column("Solver Manual", style="magenta")
            for solver in licensed_solvers:
                table.add_row(solver, SOLVER_PAGE_MAP[solver])

            rich.print(table)


@app.command(short_help="Shows the license content.")
def license(
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Shows more information about the license.",
    ),
) -> None:
    try:
        import gamspy_base
    except ModuleNotFoundError as e:
        raise ValidationError(
            "You must install gamspy_base to use this command!"
        ) from e

    license_path = utils._get_license_path(gamspy_base.directory)
    rich.print(f"[bold]License found at[/bold]: {license_path}\n")
    rich.print("[bold]License Content[/bold]")
    print("=" * 15)
    with open(license_path, encoding="utf-8") as license_file:
        lines = [line.strip() for line in license_file.readlines()]

    print("\n".join(lines))

    print_expiration_date(lines)
    print_licensed_solvers(lines, verbose)


@app.command(short_help="Shows the path of gamspy_base.")
def base():
    try:
        import gamspy_base
    except ModuleNotFoundError as e:
        raise ValidationError(
            "You must install gamspy_base to use this command!"
        ) from e

    print(gamspy_base.directory)


if __name__ == "__main__":
    app()



from __future__ import annotations

from typing import TYPE_CHECKING

from gamspy.math.misc import MathOp

if TYPE_CHECKING:
    from gamspy._algebra.operable import Operable
    from gamspy._types import OperableType


def exp(x: OperableType) -> MathOp:
    """
    Exponential of ``x`` (i.e. ``e^x``)

    Parameters
    ----------
    x : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> import math
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import exp
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.8)
    >>> b = Parameter(m, "b")
    >>> b[...] = exp(a)
    >>> math.isclose(b.toValue(), 44.701184493300815)
    True

    """
    return MathOp("exp", (x,))


def log(x: OperableType) -> MathOp:
    """
    Natural logarithm of ``x`` (i.e. logarithm base ``e`` of ``x``)

    Parameters
    ----------
    x : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> import math
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import log
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.8)
    >>> b = Parameter(m, "b")
    >>> b[...] = log(a)
    >>> math.isclose(b.toValue(), 1.33500106673234)
    True

    """
    return MathOp("log", (x,))


def log_beta(x: OperableType, y: OperableType) -> MathOp:
    """
    Log beta function (i.e. ``log(B(x, y)``)

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> import math
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import log_beta
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.8)
    >>> b = Parameter(m, "b")
    >>> b[...] = log_beta(a,5)
    >>> math.isclose(b.toValue(), -5.45446741772822)
    True

    """
    return MathOp("logBeta", (x, y))


def log_gamma(x: OperableType) -> MathOp:
    """
    Log gamma function of ``x``

    Parameters
    ----------
    x : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import log_gamma
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.8)
    >>> b = Parameter(m, "b")
    >>> b[...] = log_gamma(a)

    """
    return MathOp("logGamma", (x,))


def logit(x: OperableType) -> MathOp:
    """
    Logit Transformation (i.e. ``log(x / (1 - x))``) for ``x`` in ``(0, 1)``

    Parameters
    ----------
    x : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> import math
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import logit
    >>> m = Container()
    >>> a = Parameter(m, "a", records=0.8)
    >>> b = Parameter(m, "b")
    >>> b[...] = logit(a)
    >>> math.isclose(b.toValue(), 1.3862943611198908)
    True

    """
    return MathOp("logit", (x,))


def log2(x: OperableType) -> MathOp:
    """
    Binary logarithm (i.e. logarithm base ``2`` of ``x``)

    Parameters
    ----------
    x : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> import math
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import log2
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.8)
    >>> b = Parameter(m, "b")
    >>> b[...] = log2(a)
    >>> math.isclose(b.toValue(), 1.9259994185562224)
    True

    """
    return MathOp("log2", (x,))


def log10(x: OperableType) -> MathOp:
    """
    Common logarithm (i.e. logarithm base ``10`` of ``x``)

    Parameters
    ----------
    x : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> import math
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import log10
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.8)
    >>> b = Parameter(m, "b")
    >>> b[...] = log10(a)
    >>> math.isclose(b.toValue(), 0.5797835966168101)
    True

    """
    return MathOp("log10", (x,))


def power(base: float | Operable, exponent: OperableType) -> MathOp:
    """
    Base to the exponent power (i.e. ``base ^ exponent``)

    Parameters
    ----------
    base : OperableType
    exponent : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> import math
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import power
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.8)
    >>> b = Parameter(m, "b")
    >>> b[...] = power(a, 3)
    >>> math.isclose(b.toValue(), 54.87199999999999)
    True

    """
    return MathOp("power", (base, exponent))


def cv_power(base: float, exponent: OperableType) -> MathOp:
    """
    Real power (i.e. ``base ^ exponent`` where ``base >= 0``; error for ``base < 0``)

    Parameters
    ----------
    base : float
    exponent : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import cv_power
    >>> m = Container()
    >>> a = Parameter(m, "a", records=4)
    >>> b = Parameter(m, "b")
    >>> b[...] = cv_power(3, a)

    """
    if not isinstance(base, (float, int)):
        raise ValueError("Base must be a number")

    if base < 0:
        raise ValueError("Base must be greater than or equal to 0")

    return MathOp("cvPower", (base, exponent))


def rpower(base: OperableType | Operable, exponent: OperableType):
    """
    Returns ``x^y`` for ``x > 0`` and also for ``x = 0`` and restricted values of ``y`` (Error if ``x < 0``)

    Parameters
    ----------
    base : OperableType
    exponent : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> from gamspy import Container, Parameter, Set
    >>> from gamspy.math import rpower
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 3.8)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = rpower(a[i], 3)

    """
    return MathOp("rPower", (base, exponent))


def sign_power(base: OperableType, exponent: float):
    """
    Signed power: ``sign(base) * |base|^exponent``, for ``exponent > 0``

    Parameters
    ----------
    base : OperableType
    exponent : float

    Returns
    -------
    MathOp

    Examples
    --------
    >>> from gamspy import Container, Parameter, Set
    >>> from gamspy.math import sign_power
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 3.8)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = sign_power(a[i], 5)

    """
    if not isinstance(exponent, (float, int)):
        raise ValueError("Exponent must be a number")

    if exponent <= 0:
        raise ValueError("Exponent must be greater than 0")

    return MathOp("signPower", (base, exponent))


def sllog10(x: OperableType, S: int | float = 1.0e-150) -> MathOp:
    """
    Smooth (linear) logarithm base 10

    Parameters
    ----------
    x : OperableType
    S : int | float, by default 1.0e-150

    Returns
    -------
    MathOp

    Examples
    --------
    >>> import math
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import sllog10
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.8)
    >>> b = Parameter(m, "b")
    >>> b[...] = sllog10(a)
    >>> math.isclose(b.toValue(), 0.5797835966168101)
    True

    """
    return MathOp("sllog10", (x, S))


def sqlog10(x: OperableType, S: int | float = 1.0e-150) -> MathOp:
    """
    Smooth (quadratic) logarithm base 10

    Parameters
    ----------
    x : OperableType
    S : int | float, by default 1.0e-150

    Returns
    -------
    MathOp

    Examples
    --------
    >>> import math
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import sqlog10
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.8)
    >>> b = Parameter(m, "b")
    >>> b[...] = sqlog10(a)
    >>> math.isclose(b.toValue(), 0.5797835966168101)
    True

    """
    return MathOp("sqlog10", (x, S))


def vc_power(base: OperableType, exponent: OperableType):
    """
    Returns ``x^y`` for ``x >= 0`` (error for ``x < 0``)

    Parameters
    ----------
    base : OperableType
    exponent : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import vc_power
    >>> m = Container()
    >>> a = Parameter(m, "a", records=4)
    >>> b = Parameter(m, "b")
    >>> b[...] = vc_power(a, 3)

    """
    return MathOp("vcPower", (base, exponent))


def sqr(x: OperableType) -> MathOp:
    """
    Square of ``x`` (i.e. ``x^2``)

    Parameters
    ----------
    x : OperableType

    Returns
    -------
    MathOp

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import sqr
    >>> m = Container()
    >>> a = Parameter(m, "a", records=4)
    >>> b = Parameter(m, "b")
    >>> b[...] = sqr(a)
    >>> b.toValue()
    np.float64(16.0)

    """
    return MathOp("sqr", (x,))



from __future__ import annotations

import os
import subprocess
import sys
from typing import TYPE_CHECKING, Annotated

import certifi
import typer

import gamspy.utils as utils
from gamspy.exceptions import GamspyException, ValidationError

from .util import remove_solver_entry

if TYPE_CHECKING:
    from collections.abc import Iterable

app = typer.Typer(
    rich_markup_mode="rich",
    short_help="To uninstall licenses and solvers.",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy uninstall license | gamspy uninstall solver <solver_name>",
    context_settings={"help_option_names": ["-h", "--help"]},
)


@app.command(
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy uninstall license",
    short_help="To uninstall the current license",
)
def license():
    try:
        os.unlink(os.path.join(utils.DEFAULT_DIR, "gamspy_license.txt"))
    except FileNotFoundError:
        ...


@app.command(
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy uninstall solver <solver_name>",
    short_help="To uninstall solvers",
)
def solver(
    solver: Annotated[
        list[str] | None,
        typer.Argument(
            help="solver names to be uninstalled",
            autocompletion=lambda: [
                s.lower()
                for s in utils.getInstalledSolvers(utils._get_gamspy_base_directory())
            ],
        ),
    ] = None,
    uninstall_all_solvers: bool = typer.Option(
        False,
        "--uninstall-all-solvers",
        "--all",
        "-a",
        help="Uninstalls all add-on solvers.",
    ),
    skip_pip_uninstall: bool = typer.Option(
        False,
        "--skip-pip-install",
        "-s",
        help="If you already have the solver uninstalled, skip pip uninstall and update gamspy installed solver list.",
    ),
    use_uv: bool = typer.Option(
        False, "--use-uv", help="Use uv instead of pip to uninstall solvers."
    ),
):
    try:
        import gamspy_base
    except ModuleNotFoundError as e:
        raise ValidationError(
            "You must install gamspy_base to use this command!"
        ) from e

    addons_path = os.path.join(utils.DEFAULT_DIR, "solvers.txt")
    environment_variables = os.environ.copy()
    if "CURL_CA_BUNDLE" not in environment_variables:
        environment_variables["CURL_CA_BUNDLE"] = certifi.where()

    def remove_addons(addons: Iterable[str]):
        for item in addons:
            solver_name = item.lower()

            installed_solvers = utils.getInstalledSolvers(gamspy_base.directory)
            removable_solvers = set(installed_solvers) - set(
                gamspy_base.default_solvers
            )
            if solver_name.upper() not in removable_solvers:
                raise ValidationError(
                    f'Given solver name ("{solver_name}") is not valid. Installed'
                    f" solvers that can be uninstalled: {sorted(removable_solvers)}"
                )

            if not skip_pip_uninstall:
                # uninstall specified solver
                if use_uv:
                    command = [
                        "uv",
                        "pip",
                        "uninstall",
                        f"gamspy-{solver_name}",
                    ]
                else:
                    command = [
                        sys.executable,
                        "-m",
                        "pip",
                        "uninstall",
                        f"gamspy-{solver_name}",
                        "-y",
                    ]
                try:
                    _ = subprocess.run(
                        command,
                        check=True,
                        encoding="utf-8",
                        stderr=subprocess.PIPE,
                    )
                except subprocess.CalledProcessError as e:
                    raise GamspyException(
                        f"Could not uninstall gamspy-{solver_name}: {e.output}"
                    ) from e

            # do not delete files from gamspy_base as other solvers might depend on it
            gamspy_base_dir = utils._get_gamspy_base_directory()
            remove_solver_entry(gamspy_base_dir, solver_name)

            try:
                with open(addons_path) as file:
                    installed = file.read().splitlines()
            except FileNotFoundError:
                installed = []

            try:
                installed.remove(solver_name.upper())
            except ValueError:
                ...

            with open(addons_path, "w") as file:
                file.write("\n".join(installed) + "\n")

    if uninstall_all_solvers:
        installed_solvers = utils.getInstalledSolvers(gamspy_base.directory)
        solvers = [
            solver
            for solver in installed_solvers
            if solver not in gamspy_base.default_solvers
        ]
        remove_addons(solvers)

        # All add-on solvers are gone.
        return

    if solver is None:
        raise ValidationError(
            "Solver name is missing: `gamspy uninstall solver <solver_name>`"
        )

    remove_addons(solver)


if __name__ == "__main__":
    app()



from __future__ import annotations

import importlib
import itertools
import uuid
from typing import TYPE_CHECKING, Any

import gamspy as gp
import gamspy.formulations.utils as utils
from gamspy.exceptions import ValidationError
from gamspy.formulations.ml.decision_tree_struct import DecisionTreeStruct
from gamspy.formulations.ml.regression_tree import RegressionTree

if TYPE_CHECKING:
    from sklearn.ensemble import GradientBoostingRegressor
    from sklearn.tree import DecisionTreeRegressor

    from gamspy import (
        Alias,
        Equation,
        Parameter,
        Set,
        Variable,
    )


class GradientBoosting:
    """
    Formulation generator for Gradient Boosted Trees in GAMSPy.

    Parameters
    ----------
    container : Container
        Container that will contain the new variable and equations.
    ensemble: GradientBoostingRegressor | list[DecisionTreeStruct]
        - A fitted `sklearn.ensemble.GradientBoostingRegressor` instance,
        - If `sklearn.ensemble.GradientBoostingRegressor` is not utilized, the ensembled trees information
          can be supplied via a list of `DecisionTreeStruct` dataclasse instances, which represents
          the same components as those in `sklearn.tree`.
          See :meth:`DecisionTreeStruct <gamspy.formulations.DecisionTreeStruct>` for details on required attributes.

    name_prefix : str | None
        Prefix for generated GAMSPy symbols, by default None which means
        random prefix. Using the same name_prefix in different formulations causes name
        conflicts. Do not use the same name_prefix again.

    bias: float | 1
        Bias term used to consolidate the final output using the contribution of each tree. This is generally
        the average of the output data used for training and it is useful when `ensemble` is a `list[DecisionTreeStruct]`.
        Otherwise, this is deduced from `ensemble` itself.

    learning_rate: float | 0.1
        Rate at which each tree's contribution is reduced, by default is 0.1. This is useful when `ensemble`
        is a `list[DecisionTreeStruct]`. Otherwise, this is deduced from `ensemble` itself.

    Examples
    --------
    >>> import gamspy as gp
    >>> import numpy as np
    >>> from gamspy.math import dim
    >>> np.random.seed(42)
    >>> m = gp.Container()
    >>> in_data = np.random.randint(0, 10, size=(5, 2))
    >>> out_data = np.random.randint(1, 3, size=(5, 1))
    >>> tree1_attribute = {
    ...     "capacity": 3,
    ...     "children_left": np.array([1, -1, -1]),
    ...     "children_right": np.array([2, -1, -1]),
    ...     "feature": np.array([0, -2, -2]),
    ...     "n_features": 2,
    ...     "threshold": np.array([4.0, -2.0, -2.0]),
    ...     "value": np.array([[-4.4408921e-17], [-8.0000000e-01], [2.0000000e-01]]),
    ... }
    >>> tree2_attribute = {
    ...     "capacity": 3,
    ...     "children_left": np.array([1, -1, -1]),
    ...     "children_right": np.array([2, -1, -1]),
    ...     "feature": np.array([0, -2, -2]),
    ...     "n_features": 2,
    ...     "threshold": np.array([4.0, -2.0, -2.0]),
    ...     "value": np.array([[-8.8817842e-17], [-6.4000000e-01], [1.6000000e-01]]),
    ... }
    >>> gb_trees = [gp.formulations.DecisionTreeStruct(**tree1_attribute), gp.formulations.DecisionTreeStruct(**tree2_attribute)]
    >>> dt_model = gp.formulations.GradientBoosting(m, gb_trees)
    >>> x = gp.Variable(m, "x", domain=dim((5, 2)), type="positive")
    >>> x.up[:, :] = 10
    >>> y, eqns = dt_model(x)
    >>> set_of_samples = y.domain[0]
    >>> set_of_samples.name
    'DenseDim5_1'

    """

    def __init__(
        self,
        container: gp.Container,
        ensemble: GradientBoostingRegressor | list[DecisionTreeStruct],
        name_prefix: str | None = None,
        bias: float = 1,
        learning_rate: float = 0.1,
    ):
        if not isinstance(container, gp.Container):
            raise ValidationError(f"{container} is not a gp.Container.")

        def _validate_ensemble(
            ensemble,
        ) -> list[DecisionTreeRegressor | DecisionTreeStruct]:
            if isinstance(ensemble, list) and all(
                isinstance(item, DecisionTreeStruct) for item in ensemble
            ):
                return ensemble
            else:
                try:
                    sklearn_boosting = importlib.import_module("sklearn.ensemble")
                    if isinstance(ensemble, sklearn_boosting.GradientBoostingRegressor):
                        if not hasattr(ensemble, "estimators_"):
                            raise ValidationError(
                                f"{ensemble} must be a trained/fitted instance of >sklearn.ensemble.GradientBoostingRegressor<."
                            )
                        return ensemble.estimators_.flatten()
                    else:
                        raise ValidationError(
                            f"{ensemble} must be an instance of either >sklearn.ensemble.GradientBoostingRegressor< or a list of >DecisionTreeStruct<"
                        )
                except ModuleNotFoundError:
                    raise ValidationError(
                        ">sklearn.ensemble< module not found."
                    ) from None

        self.container = container

        if name_prefix is None:
            name_prefix = str(uuid.uuid4()).split("-")[0]

        self._name_prefix = name_prefix

        self._list_of_trees: list[RegressionTree] = []

        for tree in _validate_ensemble(ensemble):
            rt_instance = RegressionTree(
                self.container,
                regressor=tree,
                name_prefix=self._name_prefix,
            )
            self._list_of_trees.append(rt_instance)

        if isinstance(ensemble, list):
            # list[DecisionTreeStruct]
            self._bias = bias
            self._learning_rate = learning_rate
        else:  # GradientBoostingRegressor
            self._bias = ensemble.init_.constant_.flatten()[0]
            self._learning_rate = ensemble.learning_rate

    def __call__(
        self,
        input: gp.Parameter | gp.Variable,
        M: float | None = None,
    ) -> tuple[gp.Variable, list[gp.Equation]]:
        gb_out_list: list[gp.Variable] = []
        gb_eqn_list: list[gp.Equation] = []

        set_records_total: dict[Set | Alias | Parameter | Variable | Equation, Any] = {}

        results = (
            regression_tree._yield_call(input, M)
            for regression_tree in self._list_of_trees
        )

        zipped_results = zip(*results, strict=False)
        dt_outs = next(zipped_results)
        gb_out_list.extend(dt_outs)

        set_records_iter = next(zipped_results)
        set_of_output_dim = None
        for item, set_records_dict in set_records_iter:
            set_records_total.update(set_records_dict)
            set_of_output_dim = item

        self.container.setRecords(set_records_total)

        gb_eqn_list = list(itertools.chain.from_iterable(next(zipped_results)))

        set_of_samples = input.domain[0]

        out = gp.Variable._constructor_bypass(
            self.container,
            name=utils._generate_name("v", self._name_prefix, "real_output"),
            domain=[set_of_samples, set_of_output_dim],  # type: ignore
        )

        gb_eqn = gp.Equation._constructor_bypass(
            self.container,
            name=utils._generate_name("e", self._name_prefix, "gb_eqn"),
            domain=[set_of_samples, set_of_output_dim],  # type: ignore
            description="predicted out should be equal to the sum of gradient descent out times the learning rate.",
        )

        self.container._synch_with_gams(gams_to_gamspy=True)
        gb_eqn[...] = self._bias + self._learning_rate * sum(gb_out_list) == out
        gb_eqn_list.append(gb_eqn)

        return out, gb_eqn_list



from __future__ import annotations

import os
import shutil
import tempfile
import weakref

from gamspy.exceptions import ValidationError

DEBUGGING_LEVELS = ("delete", "keep_on_error", "keep")


def validate_arguments(
    working_directory: str | None,
    debugging_level: str,
):
    # Validate working_directory
    if working_directory == "":
        raise ValidationError("`working_directory` cannot be an empty string.")

    # Validate debug_level
    if not isinstance(debugging_level, str) or debugging_level not in DEBUGGING_LEVELS:
        raise ValidationError(f"debugging level must be one of {DEBUGGING_LEVELS}")


class Workspace:
    def __init__(
        self,
        debugging_level: str,
        working_directory: str | os.PathLike | None = None,
    ):
        if isinstance(working_directory, os.PathLike):
            working_directory = os.fspath(working_directory)

        validate_arguments(working_directory, debugging_level)

        self.debugging_level = debugging_level
        self.using_tmp_working_dir = False
        self._errors: list[str] = []

        if working_directory is None:
            self.using_tmp_working_dir = True
            self.working_directory = tempfile.mkdtemp()
        else:
            self.working_directory = os.path.abspath(working_directory)
            os.makedirs(self.working_directory, exist_ok=True)

        weakref.finalize(
            self,
            self.cleanup,
            self.using_tmp_working_dir,
            self.debugging_level,
            self.working_directory,
            self._errors,
        )

    @staticmethod
    def cleanup(
        using_tmp_working_dir: bool,
        debugging_level: str,
        working_directory: str,
        errors: list[str],
    ):
        if using_tmp_working_dir:
            try:  # in case working directory has already been deleted.
                if debugging_level == "delete":
                    shutil.rmtree(working_directory)

                if debugging_level == "keep_on_error" and len(errors) == 0:
                    shutil.rmtree(working_directory)
            except (FileNotFoundError, PermissionError):
                ...



"""
This module provides a wrapper around GDX tools of GAMS to expose them in GAMSPy CLI.
Run `gamspy gdx -h` to see available gdx tools in GAMSPy CLI.
"""

from __future__ import annotations

import os
import platform
import subprocess
from typing import Annotated

import gamspy_base
import typer

app = typer.Typer(
    rich_markup_mode="rich",
    short_help="To dump and compare GDX files.",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy gdx dump gdxfile.gdx | gamspy gdx diff gdxfile.gdx gdxfile2.gdx",
    context_settings={"help_option_names": ["-h", "--help"]},
)

VALID_DELIMS = {"period", "comma", "tab", "blank", "semicolon"}
VALID_DECIMAL_SEP = {"period", "comma"}
VALID_FORMATS = {"normal", "gamsbas", "csv"}
VALID_DFORMATS = {"normal", "hexponential", "hexBytes"}
VALID_YN = {"Y", "N"}
VALID_FIELDS = {"L", "M", "Up", "Lo", "Prior", "Scale", "All"}
VALID_SETDESC = {"Y", "N"}


def complete_delim(ctx: typer.Context, incomplete: str):
    return [d for d in VALID_DELIMS if d.startswith(incomplete)]


def complete_decimalsep(ctx: typer.Context, incomplete: str):
    return [s for s in VALID_DECIMAL_SEP if s.startswith(incomplete)]


def complete_format(ctx: typer.Context, incomplete: str):
    return [f for f in VALID_FORMATS if f.startswith(incomplete)]


def complete_dformat(ctx: typer.Context, incomplete: str):
    return [f for f in VALID_DFORMATS if f.startswith(incomplete)]


def complete_yes_no(ctx: typer.Context, incomplete: str):
    return [v for v in VALID_YN if v.startswith(incomplete)]


def complete_field(ctx: typer.Context, incomplete: str):
    return [f for f in VALID_FIELDS if f.startswith(incomplete)]


@app.command(
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy gdx dump gdxfile.gdx",
    short_help="To dump the content of a GDX file in ASCII format.",
)
def dump(
    filename: str = typer.Argument(..., help="Input GDX filename"),
    version: bool = typer.Option(
        False, "--version", "-v", help="Write version info of input file only"
    ),
    output: str | None = typer.Option(
        None, "--output", "-o", help="Write output to file"
    ),
    symb: str | None = typer.Option(
        None, "--symb", "-s", help="Select a single identifier"
    ),
    ueltable: str | None = typer.Option(
        None, "--ueltable", "-u", help="Include all unique elements"
    ),
    delim: str | None = typer.Option(
        None,
        "--delim",
        "-d",
        help="Dimension delimiter",
        autocompletion=complete_delim,
    ),
    decimalsep: str | None = typer.Option(
        None,
        "--decimalsep",
        "-p",
        help="Decimal separator",
        autocompletion=complete_decimalsep,
    ),
    noheader: bool = typer.Option(
        False, "--noheader", help="Suppress writing of headers"
    ),
    nodata: bool = typer.Option(False, "--nodata", help="Write headers only; no data"),
    csvallfields: bool = typer.Option(
        False,
        "--csvallfields",
        help="Write all variable/equation fields in CSV",
    ),
    csvsettext: bool = typer.Option(
        False, "--csvsettext", help="Write set element text in CSV"
    ),
    symbols: bool = typer.Option(
        False, "--symbols", "-S", help="Get list of all symbols"
    ),
    domaininfo: bool = typer.Option(
        False, "--domaininfo", help="Show domain information"
    ),
    symbolsasset: bool = typer.Option(False, "--symbolsasset", help="Symbols as set"),
    symbolsassetdi: bool = typer.Option(
        False, "--symbolsassetdi", help="Symbols as set incl. domain info"
    ),
    settext: bool = typer.Option(False, "--settext", help="Show set text"),
    format: str | None = typer.Option(
        None,
        "--format",
        "-f",
        help="Output format: normal, gamsbas, csv",
        autocompletion=complete_format,
    ),
    dformat: str | None = typer.Option(
        None,
        "--dformat",
        "-F",
        help="Data format: normal, hexponential, hexBytes",
        autocompletion=complete_dformat,
    ),
    cdim: str | None = typer.Option(
        None,
        "--cdim",
        help="Use last dim as column headers (Y/N)",
        autocompletion=complete_yes_no,
    ),
    filterdef: str | None = typer.Option(
        None,
        "--filterdef",
        "-x",
        help="Filter default values (Y/N)",
        autocompletion=complete_yes_no,
    ),
    epsout: str | None = typer.Option(None, "--epsout", help="String for EPS"),
    naout: str | None = typer.Option(None, "--naout", help="String for NA"),
    pinfout: str | None = typer.Option(None, "--pinfout", help="String for +Inf"),
    minfout: str | None = typer.Option(None, "--minfout", help="String for -Inf"),
    undfout: str | None = typer.Option(None, "--undfout", help="String for Undefined"),
    zeroout: str | None = typer.Option(None, "--zeroout", help="String for Zero"),
    header: str | None = typer.Option(
        None, "--header", help="New header for CSV output"
    ),
):
    GDXDUMP_PATH = os.path.join(gamspy_base.directory, "gdxdump")
    if platform.system() == "Windows":
        GDXDUMP_PATH = f"{GDXDUMP_PATH}.exe"

    if not filename.endswith(".gdx"):
        filename += ".gdx"

    if not os.path.exists(filename):
        typer.echo(f"File not found: {filename}", err=True)
        raise typer.Exit(code=1)

    # Validation
    if delim and delim not in VALID_DELIMS:
        typer.echo(
            f"Invalid Delim: {delim}. Must be one of: {', '.join(VALID_DELIMS)}",
            err=True,
        )
        raise typer.Exit(code=1)
    if decimalsep and decimalsep not in VALID_DECIMAL_SEP:
        typer.echo(
            f"Invalid DecimalSep: {decimalsep}. Must be one of: {', '.join(VALID_DECIMAL_SEP)}",
            err=True,
        )
        raise typer.Exit(code=1)
    if format and format not in VALID_FORMATS:
        typer.echo(
            f"Invalid Format: {format}. Must be one of: {', '.join(VALID_FORMATS)}",
            err=True,
        )
        raise typer.Exit(code=1)
    if dformat and dformat not in VALID_DFORMATS:
        typer.echo(
            f"Invalid dFormat: {dformat}. Must be one of: {', '.join(VALID_DFORMATS)}",
            err=True,
        )
        raise typer.Exit(code=1)
    if cdim and cdim not in VALID_YN:
        typer.echo(f"Invalid CDim: {cdim}. Must be Y or N", err=True)
        raise typer.Exit(code=1)
    if filterdef and filterdef not in VALID_YN:
        typer.echo(f"Invalid FilterDef: {filterdef}. Must be Y or N", err=True)
        raise typer.Exit(code=1)

    cmd = [GDXDUMP_PATH, filename]
    if version:
        cmd.append("-V")
    if output:
        cmd.append(f"Output={output}")
    if symb:
        cmd.append(f"Symb={symb}")
    if ueltable:
        cmd.append(f"UelTable={ueltable}")
    if delim:
        cmd.append(f"Delim={delim}")
    if decimalsep:
        cmd.append(f"DecimalSep={decimalsep}")
    if noheader:
        cmd.append("NoHeader")
    if nodata:
        cmd.append("NoData")
    if csvallfields:
        cmd.append("CSVAllFields")
    if csvsettext:
        cmd.append("CSVSetText")
    if symbols:
        cmd.append("Symbols")
    if domaininfo:
        cmd.append("DomainInfo")
    if symbolsasset:
        cmd.append("SymbolsAsSet")
    if symbolsassetdi:
        cmd.append("SymbolsAsSetDI")
    if settext:
        cmd.append("SetText")
    if format:
        cmd.append(f"Format={format}")
    if dformat:
        cmd.append(f"dFormat={dformat}")
    if cdim:
        cmd.append(f"CDim={cdim}")
    if filterdef:
        cmd.append(f"FilterDef={filterdef}")
    if epsout:
        cmd.append(f"EpsOut={epsout}")
    if naout:
        cmd.append(f"NaOut={naout}")
    if pinfout:
        cmd.append(f"PinfOut={pinfout}")
    if minfout:
        cmd.append(f"MinfOut={minfout}")
    if undfout:
        cmd.append(f"UndfOut={undfout}")
    if zeroout:
        cmd.append(f"ZeroOut={zeroout}")
    if header:
        cmd.append(f"Header={header}")

    result = subprocess.run(cmd, text=True)
    raise typer.Exit(code=result.returncode)


@app.command(
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy gdx diff gdxfile.gdx gdxfile2.gdx",
    short_help="To show the differences between two GDX files.",
)
def diff(
    file1: str = typer.Argument(..., help="First input GDX file"),
    file2: str = typer.Argument(..., help="Second input GDX file"),
    diffile: str | None = typer.Argument(
        None, help="Optional output GDX file for differences"
    ),
    eps: float | None = typer.Option(
        None, "--eps", "-e", help="Epsilon for comparison"
    ),
    releps: float | None = typer.Option(
        None, "--releps", "-r", help="Relative epsilon for comparison"
    ),
    field: str | None = typer.Option(
        None,
        "--field",
        "-f",
        help="Field to compare: L, M, Up, Lo, Prior, Scale, All",
        autocompletion=complete_field,
    ),
    fldonly: bool = typer.Option(
        False, "--fldonly", "-o", help="Write only selected field"
    ),
    diffonly: bool = typer.Option(
        False,
        "--diffonly",
        "-d",
        help="Write only differences with field dimension",
    ),
    cmpdefaults: bool = typer.Option(
        False, "--cmpdefaults", "-c", help="Compare default values"
    ),
    cmpdomains: bool = typer.Option(
        False, "--cmpdomains", "-m", help="Compare domain information"
    ),
    matrixfile: bool = typer.Option(
        False, "--matrixfile", "-x", help="Compare GAMS matrix files"
    ),
    ignoreorder: bool = typer.Option(
        False, "--ignoreorder", help="Ignore UEL order of input files"
    ),
    setdesc: str | None = typer.Option(
        None,
        "--setdesc",
        help="Compare set element descriptions (Y/N)",
        autocompletion=complete_yes_no,
    ),
    id: Annotated[
        list[str] | None,
        typer.Option("--id", "-i", help="One or more identifiers to include"),
    ] = None,
    skipid: Annotated[
        list[str] | None,
        typer.Option("--skipid", "-s", help="One or more identifiers to skip"),
    ] = None,
):
    GDXDIFF_PATH = os.path.join(gamspy_base.directory, "gdxdiff")
    if platform.system() == "Windows":
        GDXDIFF_PATH = f"{GDXDIFF_PATH}.exe"

    if not file1.endswith(".gdx"):
        file1 += ".gdx"

    if not os.path.exists(file1):
        typer.echo(f"File not found: {file1}", err=True)
        raise typer.Exit(code=1)

    if not file2.endswith(".gdx"):
        file2 += ".gdx"

    if not os.path.exists(file2):
        typer.echo(f"File not found: {file2}", err=True)
        raise typer.Exit(code=1)

    if diffile and not diffile.endswith(".gdx"):
        diffile += ".gdx"

    # Validate restricted strings
    if field and field not in VALID_FIELDS:
        typer.echo(
            f"Invalid field: '{field}'. Must be one of: {', '.join(VALID_FIELDS)}",
            err=True,
        )
        raise typer.Exit(code=1)

    if setdesc and setdesc not in VALID_SETDESC:
        typer.echo(f"Invalid SetDesc value: '{setdesc}'. Must be Y or N", err=True)
        raise typer.Exit(code=1)

    # Build the command
    cmd = [GDXDIFF_PATH, file1, file2]
    if diffile:
        cmd.append(diffile)
    if eps is not None:
        cmd.append(f"Eps={eps}")
    if releps is not None:
        cmd.append(f"RelEps={releps}")
    if field:
        cmd.append(f"Field={field}")
    if fldonly:
        cmd.append("FldOnly")
    if diffonly:
        cmd.append("DiffOnly")
    if cmpdefaults:
        cmd.append("CmpDefaults")
    if cmpdomains:
        cmd.append("CmpDomains")
    if matrixfile:
        cmd.append("MatrixFile")
    if ignoreorder:
        cmd.append("IgnoreOrder")
    if setdesc:
        cmd.append(f"SetDesc={setdesc}")
    if id:
        for i in id:
            cmd.append(f"ID={i}")
    if skipid:
        for s in skipid:
            cmd.append(f"SkipID={s}")

    result = subprocess.run(cmd, text=True)
    raise typer.Exit(code=result.returncode)


if __name__ == "__main__":
    app()



from __future__ import annotations

import math

import gamspy as gp
import gamspy.formulations.utils as utils
from gamspy.exceptions import ValidationError


def _get_new_domain(
    x: gp.Variable | gp.Parameter, dims: list[int]
) -> tuple[list[gp.Set | gp.Alias], gp.Set]:
    lens = [len(x.domain[d]) for d in dims]
    new_card = math.prod(lens)

    # get new domain
    flattened = gp.math._generate_dims(x.container, [new_card])
    # flattened should be unique, this is only for declaration
    flattened = utils._next_domains(
        flattened, [*x.domain[: min(dims)], *x.domain[max(dims) + 1 :]]
    )[0]
    new_domain = []
    new_domain.extend(x.domain[: min(dims)])
    new_domain.append(flattened)
    new_domain.extend(x.domain[max(dims) + 1 :])

    return new_domain, flattened


def _flatten_dims_par(
    x: gp.Parameter, dims: list[int]
) -> tuple[gp.Parameter, list[gp.Equation]]:
    data = x.toDense()

    m = x.container
    new_domain, _ = _get_new_domain(x, dims)

    new_shape = [len(d) for d in new_domain]

    if data is not None:
        data = data.reshape(new_shape)

    out = m.addParameter(domain=new_domain, records=data)
    return out, []


def _generate_index_matching_statement(
    domains: list[gp.Set], flattened: gp.Set, matching_set: gp.Set
) -> str:
    base_txt = "option {}({}:{})"
    domains_str = ",".join([x.name for x in domains])
    return base_txt.format(matching_set.name, domains_str, flattened.name)


def _propagate_bounds(x, out):
    """Propagate bounds from the input to the output variable"""

    m = x.container

    # set domain for variable
    x_domain = x.domain
    x_domain = utils._next_domains(x_domain, [])

    bounds_set = m.addSet(records=["lb", "ub"])
    bounds = m.addParameter(domain=[bounds_set, *x_domain])

    # capture original bounds
    bounds[("lb",) + tuple(x_domain)] = x.lo[x_domain]
    bounds[("ub",) + tuple(x_domain)] = x.up[x_domain]

    # reshape bounds based on the output variable's shape
    # when bounds.records is None, it means the bounds are zeros
    nb_data = (
        None if bounds.records is None else bounds.toDense().reshape((2,) + out.shape)
    )

    # set new domain for bounds
    nb_dom = [bounds_set, *out.domain]

    new_bounds = m.addParameter(domain=nb_dom, records=nb_data)

    # assign new bounds to the output variable
    out.lo[...] = new_bounds[("lb",) + tuple(out.domain)]
    out.up[...] = new_bounds[("ub",) + tuple(out.domain)]


def _flatten_dims_var(
    x: gp.Variable, dims: list[int], propagate_bounds: bool = True
) -> tuple[gp.Variable, list[gp.Equation]]:
    m = x.container
    new_domain, flattened = _get_new_domain(x, dims)

    out = m.addVariable(
        domain=new_domain
    )  # outputs domain nearly matches the input domain

    if propagate_bounds and x.records is not None:
        _propagate_bounds(x, out)

    # match the flattened set to correct dims
    forwarded_domain = utils._next_domains([flattened, *x.domain], [])
    doms_to_flatten = [forwarded_domain[d + 1] for d in dims]

    name = "ds_" + gp.utils._get_unique_name()
    subset = m.addSet(name, domain=[*doms_to_flatten, flattened])
    m.addGamsCode(
        _generate_index_matching_statement(doms_to_flatten, flattened, subset)
    )

    set_out = m.addEquation(domain=forwarded_domain)
    set_out[forwarded_domain].where[subset[[*doms_to_flatten, flattened]]] = (
        out[
            [
                *forwarded_domain[1 : min(dims) + 1],
                flattened,
                *forwarded_domain[max(dims) + 2 :],
            ]
        ]
        == x[forwarded_domain[1:]]
    )
    return out, [set_out]


def flatten_dims(
    x: gp.Variable | gp.Parameter,
    dims: list[int],
    propagate_bounds: bool = True,
) -> tuple[gp.Parameter | gp.Variable, list[gp.Equation]]:
    """
    Flatten domains indicated by `dims` into a single domain.
    If `propagate_bounds` is True, and `x` is of type variable,
    the bounds of the input variable are propagated to the output.

    Parameters
    ----------
    x : gp.Variable | gp.Parameter
        Input to be flattened
    dims: list[int]
        List of integers indicating indices of the domains
        to be flattened. Must be consecutive indices.
    propagate_bounds: bool, optional
        Propagate bounds from the input to the output variable. Default is True.

    Examples
    --------
    >>> import gamspy as gp
    >>> from gamspy.math import dim
    >>> m = gp.Container()
    >>> inp = gp.Variable(m, domain=dim((10, 1, 24, 24)))
    >>> out, eqs = gp.formulations.flatten_dims(inp, [2, 3])
    >>> type(out)
    <class 'gamspy._symbols.variable.Variable'>
    >>> [len(x) for x in out.domain]
    [10, 1, 576]

    """
    if not isinstance(x, (gp.Parameter, gp.Variable)):
        raise ValidationError("Expected a parameter or a variable input")

    if not isinstance(propagate_bounds, bool):
        raise ValidationError("Expected a boolean for propagate_bounds")

    if len(dims) < 2:
        raise ValidationError("Expected at least 2 items in the dim array")

    x_len = len(x.domain)
    for i, d in enumerate(dims):
        if not isinstance(d, int):
            raise ValidationError("Expected integers in the dim array")

        if d < 0 or d >= x_len:
            raise ValidationError(
                "dims must contain numbers between 0 to len(domain) - 1"
            )

        if i > 0 and dims[i - 1] != d - 1:
            raise ValidationError("Expected consecutive integers in the dim array")

    for domain in x.domain:
        if len(domain) == 0:
            raise ValidationError(
                f"domain {domain} had 0 cardinality, please populate the domain first"
            )

    if isinstance(x, gp.Parameter):
        return _flatten_dims_par(x, dims)

    return _flatten_dims_var(x, dims, propagate_bounds)



from __future__ import annotations

import os
import subprocess

import certifi
import typer

import gamspy.utils as utils
from gamspy.exceptions import ValidationError

app = typer.Typer(
    rich_markup_mode="rich",
    short_help="To retrieve a license with another node's information.",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy retrieve license <access_code> [--input <input_path>.json] [--output <output_path>.json]",
    context_settings={"help_option_names": ["-h", "--help"]},
)


@app.command(
    short_help="Retrives the license with the given node information.",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy retrieve license <access_code> [--input <input_path>.json] [--output <output_path>.json]",
)
def license(
    access_code: str = typer.Argument(..., help="Access code of the license."),
    input: str = typer.Option(
        None,
        "--input",
        "-i",
        help="Input json file path to retrieve the license based on the node information.",
    ),
    output: str = typer.Option(
        None, "--output", "-o", help="Output path for the license file."
    ),
    checkout_duration: int | None = typer.Option(
        None,
        "--checkout-duration",
        "-c",
        help="Specifies a duration in hours to checkout a session.",
    ),
) -> None:
    if input is None or not os.path.isfile(input):
        raise ValidationError(
            f"Given path `{input}` is not a json file. Please use `gamspy retrieve license <access_code> -i <json_file_path>`"
        )

    if access_code is None:
        raise ValidationError(f"Given licence id `{access_code}` is not valid!")

    environment_variables = os.environ.copy()
    if "CURL_CA_BUNDLE" not in environment_variables:
        environment_variables["CURL_CA_BUNDLE"] = certifi.where()

    gamspy_base_dir = utils._get_gamspy_base_directory()
    command = [
        os.path.join(gamspy_base_dir, "gamsgetkey"),
        access_code,
        "-i",
        input,
    ]
    if checkout_duration:
        command.append("-c")
        command.append(str(checkout_duration))

    process = subprocess.run(
        command, text=True, capture_output=True, env=environment_variables
    )

    if process.returncode:
        raise ValidationError(process.stderr)

    print(process.stdout)
    if output:
        with open(output, "w") as file:
            file.write(process.stdout)


if __name__ == "__main__":
    app()
